# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides constant values related to parametric studies."""


class ColumnNames:
    """Provides column names for the parametric study data frame.

    Values are stored internally as a :class:`Pandas DataFrame <pandas.DataFrame>`.
    Column name definitions follow.
    """

    ITERATION = "Iteration"
    """Optional iteration number, useful for tracking the sequence of
    simulation groups."""
    PRIORITY = "Priority"
    """Priority value used to determine execution order."""
    TYPE = "Type"
    """Type of simulation."""
    ID = "ID"
    """Unique identifier for the simulation."""
    STATUS = "Status"
    """Status of the simulation."""
    MATERIAL = "Material"
    """Name of material used during simulation."""
    HEATER_TEMPERATURE = "Heater Temp (C)"
    """Heater temperature (C)."""
    LAYER_THICKNESS = "Layer Thickness (m)"
    """Powder deposition layer thickness (m)."""
    BEAM_DIAMETER = "Beam Diameter (m)"
    """Laser beam diameter (m)."""
    LASER_POWER = "Laser Power (W)"
    """Laser power (W)."""
    SCAN_SPEED = "Scan Speed (m/s)"
    """Laser scan speed (m/s)."""
    PV_RATIO = "Laser Power/Scan Speed (J/m)"
    """Ratio of laser power to scan speed (J/m)."""
    START_ANGLE = "Start Angle (degrees)"
    """Hatch scan angle for first layer (degrees)."""
    ROTATION_ANGLE = "Rotation Angle (degrees)"
    """Hatch rotation angle for subsequent layers (degrees)."""
    HATCH_SPACING = "Hatch Spacing (m)"
    """Hatch spacing (m)."""
    STRIPE_WIDTH = "Stripe Width (m)"
    """Stripe width (m)."""
    ENERGY_DENSITY = "Energy Density (J/m^3)"
    """Laser power divided by build rate (J/m^3)."""
    SINGLE_BEAD_LENGTH = "Single Bead Length (m)"
    """Length of single bead to simulate (m)."""
    BUILD_RATE = "Build Rate (m^3/s)"
    """Layer thickness * scan speed * hatch spacing (m^3/s)."""
    MELT_POOL_WIDTH = "Melt Pool Width (m)"
    """Median melt pool width measured at the top of the powder layer (m)."""
    MELT_POOL_DEPTH = "Melt Pool Depth (m)"
    """Median melt pool depth measured from the top of the powder layer (m)."""
    MELT_POOL_LENGTH = "Melt Pool Length (m)"
    """Median melt pool length measured at the top of the powder layer (m)."""
    MELT_POOL_LENGTH_OVER_WIDTH = "Melt Pool Length/Width"
    """Ratio of MELT_POOL_LENGTH to MELT_POOL_WIDTH."""
    MELT_POOL_REFERENCE_WIDTH = "Melt Pool Ref Width (m)"
    """Median melt pool width measured at the top of the base plate (m)."""
    MELT_POOL_REFERENCE_DEPTH = "Melt Pool Ref Depth (m)"
    """Median melt pool depth measured from the top of the base plate (m)."""
    MELT_POOL_REFERENCE_DEPTH_OVER_WIDTH = "Melt Pool Ref Depth/Width"
    """Ratio of MELT_POOL_REFERENCE_DEPTH to MELT_POOL_REFERENCE_WIDTH."""
    POROSITY_SIZE_X = "Porosity Size X (m)"
    """X dimension size of porosity sample to simulate (m)."""
    POROSITY_SIZE_Y = "Porosity Size Y (m)"
    """Y dimension size of Porosity sample to simulate (m)."""
    POROSITY_SIZE_Z = "Porosity Size Z (m)"
    """Z dimension size of Porosity sample to simulate (m)."""
    RELATIVE_DENSITY = "Relative Density"
    """Relative density of simulated porosity sample."""
    MICRO_MIN_X = "Micro Min X (m)"
    """Minimum X dimension position of microstructure sample (m)."""
    MICRO_MIN_Y = "Micro Min Y (m)"
    """Minimum Y dimension position of microstructure sample (m)."""
    MICRO_MIN_Z = "Micro Min Z (m)"
    """Minimum Z dimension position of microstructure sample (m)."""
    MICRO_SIZE_X = "Micro Size X (m)"
    """X dimension size of microstructure sample to simulate (m)."""
    MICRO_SIZE_Y = "Micro Size Y (m)"
    """Y dimension size of microstructure sample to simulate (m)."""
    MICRO_SIZE_Z = "Micro Size Z (m)"
    """Z dimension size of microstructure sample to simulate (m)."""
    MICRO_SENSOR_DIM = "Micro Sensor Dim (m)"
    """Sensor dimension used in microstructure simulations (m)."""
    COOLING_RATE = "Cooling Rate (K/s)"
    """User-provided cooling rate used in microstructure simulations (K/s)."""
    THERMAL_GRADIENT = "Thermal Gradient (K/m)"
    """User-provided thermal gradient used in microstructure simulations (K/m)."""
    MICRO_MELT_POOL_WIDTH = "Micro Melt Pool Width (m)"
    """User-provided melt pool width used in microstructure simulation (m)."""
    MICRO_MELT_POOL_DEPTH = "Micro Melt Pool Depth (m)"
    """User-provided melt pool depth used in microstructure simulation (m)."""
    RANDOM_SEED = "Random Seed"
    """User-provided random seed used in microstructure simulation."""
    XY_AVERAGE_GRAIN_SIZE = "XY Average Grain Size (microns)"
    """Average microstructure grain size in the XY plane (microns)."""
    XZ_AVERAGE_GRAIN_SIZE = "XZ Average Grain Size (microns)"
    """Average microstructure grain size in the XZ plane (microns)."""
    YZ_AVERAGE_GRAIN_SIZE = "YZ Average Grain Size (microns)"
    """Average microstructure grain size in the YZ plane (microns)."""
    ERROR_MESSAGE = "Error Message"
    """Error message if simulation failed."""


DEFAULT_ITERATION = 0
"""Default iteration assigned to new simulations."""
DEFAULT_PRIORITY = 1
"""Default priority assigned to new simulations."""
FORMAT_VERSION = 3
"""Parametric study file format version."""



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides input and result summary containers for thermal history simulations."""

from ansys.additive.core.geometry_file import BuildFile, StlFile
from ansys.additive.core.machine import AdditiveMachine
from ansys.additive.core.material import AdditiveMaterial
from ansys.additive.core.simulation import (
    SimulationInputBase,
    SimulationStatus,
    SimulationSummaryBase,
)
from ansys.api.additive.v0.additive_domain_pb2 import BuildFile as BuildFileMessage
from ansys.api.additive.v0.additive_domain_pb2 import (
    CoaxialAverageSensorInputs as CoaxialAverageSensorInputsMessage,
)
from ansys.api.additive.v0.additive_domain_pb2 import Range as RangeMessage
from ansys.api.additive.v0.additive_domain_pb2 import StlFile as StlFileMessage
from ansys.api.additive.v0.additive_domain_pb2 import (
    ThermalHistoryInput as ThermalHistoryInputMessage,
)
from ansys.api.additive.v0.additive_simulation_pb2 import SimulationRequest


class Range:
    """Defines a range of values."""

    def __init__(self, min: float, max: float):
        """Initialize a ``Range`` object."""
        self._min = min
        self._max = max
        if self.min > self.max:
            raise ValueError("Attempted to initialize Range with min greater than max")

    def __repr__(self):
        repr = type(self).__name__ + "\n"
        for k in self.__dict__:
            repr += k.replace("_", "", 1) + ": " + str(getattr(self, k)) + "\n"
        return repr

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Range):
            return False
        return self._min == other._min and self._max == other._max

    @property
    def min(self) -> float:
        """Minimum value of range."""
        return self._min

    @property
    def max(self) -> float:
        """Maximum value of range."""
        return self._max

    def _to_range_message(self) -> RangeMessage:
        """Transform this object into a ``RangeMessage`` object to send to the server."""
        return RangeMessage(min=self.min, max=self.max)


class CoaxialAverageSensorInputs:
    """Provides descriptions for coaxial average sensors."""

    MIN_SENSOR_RADIUS = 5e-5
    """Minimum radius for the circular field of view of the sensor (m)."""
    MAX_SENSOR_RADIUS = 1.5e-2
    """Maximum radius for the circular field of view of the sensor (m)."""

    def __init__(self, radius: float = MIN_SENSOR_RADIUS, z_heights: list[Range] = None):
        """Initialize a ``CoaxialAverageSensorInputs`` object."""
        self._radius = radius
        self._z_heights = z_heights if z_heights else []
        # use setter to validate radius
        self.radius = radius

    def __repr__(self):
        repr = type(self).__name__ + "\n"
        for k in self.__dict__:
            repr += k.replace("_", "", 1) + ": " + str(getattr(self, k)) + "\n"
        return repr

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CoaxialAverageSensorInputs):
            return False
        return self._radius == other._radius and self._z_heights == other._z_heights

    @property
    def radius(self) -> float:
        """Radius of the circular field of the view of the sensor (m).

        Valid values are from the :obj:`MIN_SENSOR_RADIUS` value to the
        :obj:`MAX_SENSOR_RADIUS` value.
        """
        return self._radius

    @radius.setter
    def radius(self, value):
        if value < self.MIN_SENSOR_RADIUS or value > self.MAX_SENSOR_RADIUS:
            raise ValueError(
                f"Radius values must be from {self.MIN_SENSOR_RADIUS} and {self.MAX_SENSOR_RADIUS}."
            )
        self._radius = value

    @property
    def z_heights(self) -> list[Range]:
        """Array of ranges along the z axis of the geometry (m).

        The simulated sensor follows the scan path for each deposit layer within each
        range.
        """
        return self._z_heights

    def _to_coaxial_average_sensor_inputs_message(
        self,
    ) -> CoaxialAverageSensorInputsMessage:
        """Create a coaxial average sensor input message to send to the server
        based upon this object.
        """
        msg = CoaxialAverageSensorInputsMessage(sensor_radius=self.radius)
        for z in self.z_heights:
            msg.z_heights.append(z._to_range_message())
        return msg


class ThermalHistoryInput(SimulationInputBase):
    """Provides input parameters for microstructure simulation."""

    def __init__(
        self,
        *,
        machine: AdditiveMachine = None,
        material: AdditiveMaterial = None,
        geometry: StlFile | BuildFile = None,
        coax_ave_sensor_inputs: CoaxialAverageSensorInputs = None,
    ):
        """Initialize a ``ThermalHistoryInput`` object."""
        super().__init__()
        self._machine = machine if machine else AdditiveMachine()
        self._material = material if material else AdditiveMaterial()
        self._geometry = geometry
        self._coax_ave_sensor_inputs = (
            coax_ave_sensor_inputs if coax_ave_sensor_inputs else CoaxialAverageSensorInputs()
        )

    def __repr__(self):
        repr = type(self).__name__ + "\n"
        for k in self.__dict__:
            if k == "_machine" or k == "_material":
                repr += "\n" + k.replace("_", "", 1) + ": " + str(getattr(self, k))
            else:
                repr += k.replace("_", "", 1) + ": " + str(getattr(self, k)) + "\n"
        return repr

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ThermalHistoryInput):
            return False
        return all(getattr(self, k) == getattr(other, k) for k in self.__dict__)

    @property
    def machine(self) -> AdditiveMachine:
        """Machine parameters."""
        return self._machine

    @machine.setter
    def machine(self, value):
        self._machine = value

    @property
    def material(self) -> AdditiveMaterial:
        """Material parameters."""
        return self._material

    @material.setter
    def material(self, value):
        self._material = value

    @property
    def geometry(self) -> StlFile | BuildFile:
        """Part geometry."""
        return self._geometry

    @geometry.setter
    def geometry(self, value):
        """Set geometry."""
        if not isinstance(value, (StlFile, BuildFile)):
            raise TypeError("Geometry must be an StlFile or BuildFile.")
        self._geometry = value

    @property
    def coax_ave_sensor_inputs(self) -> CoaxialAverageSensorInputs:
        """Coaxial average sensor inputs."""
        return self._coax_ave_sensor_inputs

    @coax_ave_sensor_inputs.setter
    def coax_ave_sensor_inputs(self, value):
        if not isinstance(value, CoaxialAverageSensorInputs):
            raise TypeError(
                "Coaxial average sensor inputs must be a 'CoaxialAverageSensorInputs' object."
            )
        self._coax_ave_sensor_inputs = value

    def _to_simulation_request(self, remote_geometry_path: str) -> SimulationRequest:
        """Convert this object into a simulation request message."""

        if not self.geometry:
            raise ValueError("Attempted to create simulation request without defining geometry")
        if not remote_geometry_path or (remote_geometry_path == ""):
            raise ValueError(
                "Attempted to create simulation request with empty remote_geometry_path"
            )

        input = ThermalHistoryInputMessage(
            machine=self.machine._to_machine_message(),
            material=self.material._to_material_message(),
            coax_ave_sensor_inputs=self.coax_ave_sensor_inputs._to_coaxial_average_sensor_inputs_message(),
        )

        if isinstance(self.geometry, StlFile):
            input.stl_file.CopyFrom(StlFileMessage(name=remote_geometry_path))
        else:
            input.build_file.CopyFrom(
                BuildFileMessage(type=self.geometry.type, name=remote_geometry_path)
            )

        return SimulationRequest(id=self.id, thermal_history_input=input)


class ThermalHistorySummary(SimulationSummaryBase):
    """Summary of a thermal history simulation."""

    def __init__(
        self,
        input: ThermalHistoryInput,
        coax_ave_output_folder: str,
        logs: str,
        status: SimulationStatus = SimulationStatus.COMPLETED,
    ):
        """Initialize a ``ThermalHistorySummary`` object."""
        if not isinstance(input, ThermalHistoryInput):
            raise ValueError("Invalid input type passed to init, " + self.__class__.__name__)
        if not isinstance(logs, str):
            raise ValueError("Invalid logs type passed to init, " + self.__class__.__name__)
        super().__init__(logs, status)
        self._input = input
        self._coax_ave_output_folder = coax_ave_output_folder

    @property
    def input(self) -> ThermalHistoryInput:
        """Simulation input.

        For more information, see the :class:`ThermalHistoryInput` class.
        """
        return self._input

    @property
    def coax_ave_output_folder(self) -> str:
        """Path to the folder containing the coaxial average sensor results.

        Results consist of VTK files, one per deposit layer, containing the thermal
        history of the scan pattern.
        """
        return self._coax_ave_output_folder



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides functions for downloading sample datasets from the PyAdditive repository."""

import os
import shutil
import urllib.request
import zipfile
from http.client import HTTPMessage

from ansys.additive.core import EXAMPLES_PATH

EXAMPLES_URI = "https://github.com/ansys/example-data/raw/master"
"""URI for the example data repository."""
PARTS_FOLDER = "pyadditive/part-only"
"""Folder containing example parts."""
BUILD_FILES_FOLDER = "pyadditive/buildfiles"
"""Folder containing example build files."""
MATERIAL_TUNING_FOLDER = "pyadditive/material_tuning_input"
"""Folder containing example material tuning input files."""
CUSTOM_MATERIAL_FOLDER = "pyadditive/custom_material_data"
"""Folder containing example custom material data files."""


def delete_downloads():
    """Delete all downloaded examples to free space or update the files."""
    shutil.rmtree(EXAMPLES_PATH)
    os.makedirs(EXAMPLES_PATH)
    return True


def decompress(filename, subdir=None) -> str:
    """Decompress a ZIP file to the ``examples`` directory.

    Parameters
    ----------
    filename : str
        Name of the file.
    subdir : str, None
        Name of the subdirectory of the ``examples`` directory
        to extract the ZIP file contents to. The default is ``None``.

    Returns
    -------
    str
       Path to the decompressed contents of the ZIP file.

    """
    outdir = EXAMPLES_PATH
    zip_ref = zipfile.ZipFile(filename, "r")
    if subdir:
        outdir = os.path.join(outdir, subdir)
        os.makedirs(outdir, exist_ok=True)
    zip_ref.extractall(outdir)
    zip_ref.close()
    return outdir


def _get_file_url(filename, directory=None):
    if directory:
        return f"{EXAMPLES_URI}/{directory}/{filename}"
    return f"{EXAMPLES_URI}/{filename}"


def _retrieve_file(url, filename) -> tuple[str, HTTPMessage]:
    """Retrieve an example data file from either a URL or local storage.

    Parameters
    ----------
    url : str
       URL for the example data file.
    filename : str
       Name for the example data file.

    Returns
    -------
    str: Local path file was downloaded to.
    HttpMessage: HTTP status message, if any.

    """
    # First check if file has already been downloaded
    local_path = os.path.join(EXAMPLES_PATH, os.path.basename(filename))
    if os.path.isfile(local_path) or os.path.isdir(local_path):
        return local_path, None

    # grab the correct url retriever
    urlretrieve = urllib.request.urlretrieve

    # Perform download
    saved_file, resp = urlretrieve(url)
    shutil.move(saved_file, local_path)
    return local_path, resp


def _download_file(filename, directory=None) -> tuple[str, HTTPMessage]:
    url = _get_file_url(filename, directory)
    return _retrieve_file(url, filename)


def download_10mm_cube():
    """Download an STL file describing a 10-millimeter cube."""
    return _download_file("10mm_cube.stl", PARTS_FOLDER)[0]


def download_small_wedge_slm_build_file():
    """Download an SLM build file for a small wedge part."""
    return _download_file("small-wedge-slm.zip", BUILD_FILES_FOLDER)[0]


class MaterialTuningExampleInputFiles:
    """Container for the example material tuning process inputs.

    Parameters
    ----------
    experiment_data_file : str
        Path to the experiment data file (CSV).
    material_configuration_file : str
        Path to the material configuration file (JSON).
    thermal_properties_lookup_file : str
        Path to the thermal properties lookup file (CSV).
    characteristic_width_lookup_file : str
        Path to the characteristic width lookup file (CSV).

    """

    def __init__(
        self,
        experiment_data_file: str,
        material_configuration_file: str,
        thermal_properties_lookup_file: str,
        characteristic_width_lookup_file: str,
    ):
        """Initialize a ``MaterialTuningInputs`` object."""
        self._experiment_data_file = experiment_data_file
        self._material_configuration_file = material_configuration_file
        self._thermal_properties_lookup_file = thermal_properties_lookup_file
        self._characteristic_width_lookup_file = characteristic_width_lookup_file

    @property
    def experiment_data_file(self):
        """Path to the experiment data file (CSV).)"""
        return self._experiment_data_file

    @property
    def material_configuration_file(self):
        """Path to the material configuration file (JSON)."""
        return self._material_configuration_file

    @property
    def thermal_properties_lookup_file(self):
        """Path to the thermal properties lookup file (CSV)."""
        return self._thermal_properties_lookup_file

    @property
    def characteristic_width_lookup_file(self):
        """Path to the characteristic width lookup file (CSV)."""
        return self._characteristic_width_lookup_file


def download_material_tuning_input() -> MaterialTuningExampleInputFiles:
    """Download the input files for the material tuning example."""
    zip_file = _download_file("material_tuning_input.zip", MATERIAL_TUNING_FOLDER)[0]
    extract_dir = decompress(zip_file, "material_tuning_input")
    experiment_data_file = os.path.join(extract_dir, "experiment_data.csv")
    if not os.path.isfile(experiment_data_file):
        raise FileNotFoundError("Failed to download experiment data file")
    material_configuration_file = os.path.join(extract_dir, "material_parameters.json")
    if not os.path.isfile(material_configuration_file):
        raise FileNotFoundError("Failed to download material configuration file")
    thermal_lookup_file = os.path.join(extract_dir, "thermal_lookup.csv")
    if not os.path.isfile(thermal_lookup_file):
        raise FileNotFoundError("Failed to download thermal lookup file")
    characteristic_width_lookup_file = os.path.join(extract_dir, "characteristic_width_lookup.csv")
    if not os.path.isfile(characteristic_width_lookup_file):
        raise FileNotFoundError("Failed to download characteristic width lookup file")
    return MaterialTuningExampleInputFiles(
        experiment_data_file,
        material_configuration_file,
        thermal_lookup_file,
        characteristic_width_lookup_file,
    )


class CustomMaterialExampleFiles:
    """Holds the files associated with a custom material definition.

    Parameters
    ----------
    material_configuration_file : str
        Path to the material configuration file.
    thermal_properties_lookup_file : str
        Path to the thermal properties lookup file.
    characteristic_width_lookup_file : str
        Path to the characteristic width lookup file.

    """

    def __init__(
        self,
        material_configuration_file: str,
        thermal_properties_lookup_file: str,
        characteristic_width_lookup_file: str,
    ):
        """Initialize a ``CustomMaterialExampleFiles`` object."""

        self._material_configuration_file = material_configuration_file
        self._thermal_properties_lookup_file = thermal_properties_lookup_file
        self._characteristic_width_lookup_file = characteristic_width_lookup_file

    @property
    def material_configuration_file(self):
        """Path to the material configuration file."""
        return self._material_configuration_file

    @property
    def thermal_properties_lookup_file(self):
        """Path to the thermal properties lookup file."""
        return self._thermal_properties_lookup_file

    @property
    def characteristic_width_lookup_file(self):
        """Path to the characteristic width lookup file."""
        return self._characteristic_width_lookup_file


def download_custom_material() -> CustomMaterialExampleFiles:
    """Download the files describing a custom material."""
    zip_file = _download_file("custom_material_data.zip", CUSTOM_MATERIAL_FOLDER)[0]
    extract_dir = decompress(zip_file, "custom_material_data")
    material_configuration_file = os.path.join(extract_dir, "material_parameters.json")
    if not os.path.isfile(material_configuration_file):
        raise FileNotFoundError("Failed to download material configuration file")
    thermal_lookup_file = os.path.join(extract_dir, "thermal_lookup.csv")
    if not os.path.isfile(thermal_lookup_file):
        raise FileNotFoundError("Failed to download thermal lookup file")
    characteristic_width_lookup_file = os.path.join(extract_dir, "characteristic_width_lookup.csv")
    if not os.path.isfile(characteristic_width_lookup_file):
        raise FileNotFoundError("Failed to download characteristic width lookup file")
    return CustomMaterialExampleFiles(
        material_configuration_file,
        thermal_lookup_file,
        characteristic_width_lookup_file,
    )



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides input and result summary containers for microstructure simulations."""

import math
import os

import numpy as np
import pandas as pd
from google.protobuf.internal.containers import RepeatedCompositeFieldContainer

from ansys.additive.core import misc
from ansys.additive.core.machine import AdditiveMachine
from ansys.additive.core.material import AdditiveMaterial
from ansys.additive.core.simulation import (
    SimulationInputBase,
    SimulationStatus,
    SimulationSummaryBase,
)
from ansys.api.additive.v0.additive_domain_pb2 import (
    MicrostructureInput as MicrostructureInputMessage,
)
from ansys.api.additive.v0.additive_domain_pb2 import (
    MicrostructureResult as MicrostructureResultMessage,
)
from ansys.api.additive.v0.additive_simulation_pb2 import SimulationRequest


class MicrostructureInput(SimulationInputBase):
    """Provides input parameters for microstructure simulation.

    Units are SI (m, kg, s, K) unless otherwise noted.
    """

    DEFAULT_POSITION_COORDINATE = 0
    """Default X, Y, Z, position coordinate (m)."""
    MIN_POSITION_COORDINATE = 0
    """Minimum X, Y, Z, position coordinate (m)."""
    MAX_POSITION_COORDINATE = 10
    """Maximum X, Y, Z, position coordinate (m)."""
    DEFAULT_SAMPLE_SIZE = 1.5e-3
    """Default sample size in each dimension (m)."""
    MIN_SAMPLE_SIZE = 0.001
    """Minimum sample size in each dimension (m)."""
    MAX_SAMPLE_SIZE = 0.01
    """Maximum sample size in each dimension (m)."""
    DEFAULT_SENSOR_DIMENSION = 5e-4
    """Default sensor dimension (m)."""
    MIN_SENSOR_DIMENSION = 1e-4
    """Minimum sensor dimension (m)."""
    MAX_SENSOR_DIMENSION = 1e-3
    """Maximum sensor dimension (m)."""
    MIN_XY_SIZE_CUSHION = 5e-4
    """Minimum cushion between sensor dimension and sample size in the X and Y dimensions (m)."""
    MIN_Z_SIZE_CUSHION = 1e-3
    """Minimum cushion between sensor dimension and sample size in the Z dimension (m)."""
    DEFAULT_USE_PROVIDED_THERMAL_PARAMETERS = False
    """Default flag value indicating whether to use user-provided thermal parameters."""
    DEFAULT_COOLING_RATE = 1e6
    """Default cooling rate (K/s)."""
    MIN_COOLING_RATE = 1e5
    """Minimum cooling rate (K/s)."""
    MAX_COOLING_RATE = 1e7
    """Maximum cooling rate (K/s)."""
    DEFAULT_THERMAL_GRADIENT = 1e7
    """Default thermal gradient (K/m)."""
    MIN_THERMAL_GRADIENT = 1e5
    """Minimum thermal gradient (K/m)."""
    MAX_THERMAL_GRADIENT = 1e8
    """Maximum thermal gradient (K/m)."""
    DEFAULT_MELT_POOL_WIDTH = 1.5e-4
    """Default melt pool width (m)."""
    MIN_MELT_POOL_WIDTH = 7.5e-5
    """Minimum melt pool width (m)."""
    MAX_MELT_POOL_WIDTH = 8e-4
    """Maximum melt pool width (m)."""
    DEFAULT_MELT_POOL_DEPTH = 1e-4
    """Default melt pool depth (m)."""
    MIN_MELT_POOL_DEPTH = 1.5e-5
    """Minimum melt pool depth (m)."""
    MAX_MELT_POOL_DEPTH = 8e-4
    """Maximum melt pool depth (m)."""
    DEFAULT_RANDOM_SEED = 0
    """The default random seed, which indicates that a random seed was not provided."""
    MIN_RANDOM_SEED = 1
    """Minimum random seed."""
    MAX_RANDOM_SEED = 2**32 - 1
    """Maximum random seed."""

    def __init__(
        self,
        *,
        sample_min_x: float = DEFAULT_POSITION_COORDINATE,
        sample_min_y: float = DEFAULT_POSITION_COORDINATE,
        sample_min_z: float = DEFAULT_POSITION_COORDINATE,
        sample_size_x: float = DEFAULT_SAMPLE_SIZE,
        sample_size_y: float = DEFAULT_SAMPLE_SIZE,
        sample_size_z: float = DEFAULT_SAMPLE_SIZE,
        sensor_dimension: float = DEFAULT_SENSOR_DIMENSION,
        use_provided_thermal_parameters: bool = DEFAULT_USE_PROVIDED_THERMAL_PARAMETERS,
        cooling_rate: float = DEFAULT_COOLING_RATE,
        thermal_gradient: float = DEFAULT_THERMAL_GRADIENT,
        melt_pool_width: float = DEFAULT_MELT_POOL_WIDTH,
        melt_pool_depth: float = DEFAULT_MELT_POOL_DEPTH,
        random_seed: int = DEFAULT_RANDOM_SEED,
        machine: AdditiveMachine = None,
        material: AdditiveMaterial = None,
    ):
        """Initialize a ``MicrostructureInput`` object."""
        super().__init__()

        # we have a circular dependency here, so we validate sensor_dimension
        # and sample_size_* then assign them without calling the setters
        self.__validate_range(
            sensor_dimension,
            self.MIN_SENSOR_DIMENSION,
            self.MAX_SENSOR_DIMENSION,
            "sensor_dimension",
        )
        self.__validate_size(
            sample_size_x, sensor_dimension, self.MIN_XY_SIZE_CUSHION, "sample_size_x"
        )
        self.__validate_size(
            sample_size_y, sensor_dimension, self.MIN_XY_SIZE_CUSHION, "sample_size_y"
        )
        self.__validate_size(
            sample_size_z, sensor_dimension, self.MIN_Z_SIZE_CUSHION, "sample_size_z"
        )
        self._sensor_dimension = sensor_dimension
        self._sample_size_x = sample_size_x
        self._sample_size_y = sample_size_y
        self._sample_size_z = sample_size_z

        # use setters for remaining properties
        self.sample_min_x = sample_min_x
        self.sample_min_y = sample_min_y
        self.sample_min_z = sample_min_z
        self.use_provided_thermal_parameters = use_provided_thermal_parameters
        self.cooling_rate = cooling_rate
        self.thermal_gradient = thermal_gradient
        self.melt_pool_width = melt_pool_width
        self.melt_pool_depth = melt_pool_depth
        self.machine = machine if machine else AdditiveMachine()
        self.material = material if material else AdditiveMaterial()
        if random_seed != self.DEFAULT_RANDOM_SEED:
            self.random_seed = random_seed
        else:
            self._random_seed = random_seed

    def __repr__(self):
        repr = type(self).__name__ + "\n"
        for k in self.__dict__:
            if k == "_machine" or k == "_material":
                repr += "\n" + k.replace("_", "", 1) + ": " + str(getattr(self, k))
            else:
                repr += k.replace("_", "", 1) + ": " + str(getattr(self, k)) + "\n"
        return repr

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, MicrostructureInput):
            return False
        return (
            self.id == __o.id
            and self.sample_min_x == __o.sample_min_x
            and self.sample_min_y == __o.sample_min_y
            and self.sample_min_z == __o.sample_min_z
            and self.sample_size_x == __o.sample_size_x
            and self.sample_size_y == __o.sample_size_y
            and self.sample_size_z == __o.sample_size_z
            and self.sensor_dimension == __o.sensor_dimension
            and self.use_provided_thermal_parameters == __o.use_provided_thermal_parameters
            and self.cooling_rate == __o.cooling_rate
            and self.thermal_gradient == __o.thermal_gradient
            and self.melt_pool_width == __o.melt_pool_width
            and self.melt_pool_depth == __o.melt_pool_depth
            and self.random_seed == __o.random_seed
            and self.machine == __o.machine
            and self.material == __o.material
        )

    @staticmethod
    def __validate_range(value, min, max, name):
        if math.isnan(value):
            raise ValueError("{} must be a number.".format(name))
        if value < min or value > max:
            raise ValueError("{} must be between {} and {}.".format(name, min, max))

    @staticmethod
    def __validate_size(size_value, sensor_value, cushion, name):
        if math.isnan(size_value):
            raise ValueError("{} must be a number.".format(name))
        if size_value - sensor_value < cushion:
            raise ValueError(
                "{} must be at least {} larger than sensor_dimension.".format(name, cushion)
            )

    @property
    def machine(self):
        """Machine-related parameters."""
        return self._machine

    @machine.setter
    def machine(self, value):
        self._machine = value

    @property
    def material(self):
        """Material used during simulation."""
        return self._material

    @material.setter
    def material(self, value):
        self._material = value

    @property
    def sample_min_x(self) -> float:
        """Minimum x coordinate of the geometry sample (m).

        Valid values are from the :obj:`MIN_POSITION_COORDINATE` value to
        the :obj:`MAX_POSITION_COORDINATE` value.
        """
        return self._sample_min_x

    @sample_min_x.setter
    def sample_min_x(self, value: float):
        self.__validate_range(
            value,
            self.MIN_POSITION_COORDINATE,
            self.MAX_POSITION_COORDINATE,
            "sample_min_x",
        )
        self._sample_min_x = value

    @property
    def sample_min_y(self) -> float:
        """Minimum y coordinate of the geometry sample (m).

        Valid values are from the :obj:`MIN_POSITION_COORDINATE` value to the
        :obj:`MAX_POSITION_COORDINATE` value.
        """
        return self._sample_min_y

    @sample_min_y.setter
    def sample_min_y(self, value: float):
        self.__validate_range(
            value,
            self.MIN_POSITION_COORDINATE,
            self.MAX_POSITION_COORDINATE,
            "sample_min_y",
        )
        self._sample_min_y = value

    @property
    def sample_min_z(self) -> float:
        """Minimum z coordinate of the geometry sample (m).

        Valid values are from the :obj:`MIN_POSITION_COORDINATE` value to the
        :obj:`MAX_POSITION_COORDINATE` value.
        """
        return self._sample_min_z

    @sample_min_z.setter
    def sample_min_z(self, value: float):
        self.__validate_range(
            value,
            self.MIN_POSITION_COORDINATE,
            self.MAX_POSITION_COORDINATE,
            "sample_min_z",
        )
        self._sample_min_z = value

    @property
    def sample_size_x(self) -> float:
        """Size of the geometry sample in the x direction (m).

        Valid values are from the :obj:`MIN_SAMPLE_SIZE` value to the
        :obj:`MAX_SAMPLE_SIZE` value.
        When setting the```sample_size_x`` parameter, the value must be greater than the
        ``sensor_dimension`` value plus the :obj:`MIN_XY_SIZE_CUSHION` value.
        """
        return self._sample_size_x

    @sample_size_x.setter
    def sample_size_x(self, value: float):
        self.__validate_range(value, self.MIN_SAMPLE_SIZE, self.MAX_SAMPLE_SIZE, "sample_size_x")
        self.__validate_size(
            value, self.sensor_dimension, self.MIN_XY_SIZE_CUSHION, "sample_size_x"
        )
        self._sample_size_x = value

    @property
    def sample_size_y(self) -> float:
        """Size of the geometry sample in the y direction (m).

        Valid values are from the :obj:`MIN_SAMPLE_SIZE` value to the
        :obj:`MAX_SAMPLE_SIZE` value.
        When setting the ``sample_size_y`` parameter, the value must be greater than the
        ``sensor_dimension`` value plus the :obj:`MIN_XY_SIZE_CUSHION` value.
        """
        return self._sample_size_y

    @sample_size_y.setter
    def sample_size_y(self, value: float):
        self.__validate_range(value, self.MIN_SAMPLE_SIZE, self.MAX_SAMPLE_SIZE, "sample_size_y")
        self.__validate_size(
            value, self.sensor_dimension, self.MIN_XY_SIZE_CUSHION, "sample_size_y"
        )
        self._sample_size_y = value

    @property
    def sample_size_z(self) -> float:
        """Size of the geometry sample in the z direction (m).

        Valid values are from the :obj:`MIN_SAMPLE_SIZE` value to the
        :obj:`MAX_SAMPLE_SIZE` value.
        When setting the ``sample_size_y`` parameter, the value must be greater than the
        ``sensor_dimension`` value plus the :obj:`MIN_XY_SIZE_CUSHION` value.
        """
        return self._sample_size_z

    @sample_size_z.setter
    def sample_size_z(self, value: float):
        self.__validate_range(value, self.MIN_SAMPLE_SIZE, self.MAX_SAMPLE_SIZE, "sample_size_z")
        self.__validate_size(value, self.sensor_dimension, self.MIN_Z_SIZE_CUSHION, "sample_size_z")
        self._sample_size_z = value

    @property
    def sensor_dimension(self) -> float:
        """Dimension of the sensor (m).

        Valid values are from the :obj:`MIN_SENSOR_DIMENSION` value to the
        :obj:`MAX_SENSOR_DIMENSION` value.
        """
        return self._sensor_dimension

    @sensor_dimension.setter
    def sensor_dimension(self, value: float):
        self.__validate_range(
            value,
            self.MIN_SENSOR_DIMENSION,
            self.MAX_SENSOR_DIMENSION,
            "sensor_dimension",
        )
        size_errors = ""
        try:
            self.__validate_size(
                self.sample_size_x, value, self.MIN_XY_SIZE_CUSHION, "sample_size_x"
            )
        except ValueError as e:
            size_errors += str(e) + "\n"
        try:
            self.__validate_size(
                self.sample_size_y, value, self.MIN_XY_SIZE_CUSHION, "sample_size_y"
            )
        except ValueError as e:
            size_errors += str(e) + "\n"
        try:
            self.__validate_size(
                self.sample_size_z, value, self.MIN_Z_SIZE_CUSHION, "sample_size_z"
            )
        except ValueError as e:
            size_errors += str(e) + "\n"
        if size_errors:
            raise ValueError(size_errors)

        self._sensor_dimension = value

    @property
    def use_provided_thermal_parameters(self) -> bool:
        """Flag indicating if the ``cooling_rate``, ``thermal_gradient``, ``melt_pool_depth``, and ``melt_pool_width`` parameters have been provided by the user.

        If the value is ``False``, these values
        will be calculated. Default is ``False``.
        """  # noqa: E501
        return self._use_provided_thermal_parameters

    @use_provided_thermal_parameters.setter
    def use_provided_thermal_parameters(self, value: bool):
        self._use_provided_thermal_parameters = value

    @property
    def cooling_rate(self) -> float:
        """Material cooling rate (K/s).

        Valid values are from the :obj:`MIN_COOLING_RATE` value to the
        :obj:`MAX_COOLING_RATE` value.
        """
        return self._cooling_rate

    @cooling_rate.setter
    def cooling_rate(self, value: float):
        self.__validate_range(value, self.MIN_COOLING_RATE, self.MAX_COOLING_RATE, "cooling_rate")
        self._cooling_rate = value

    @property
    def thermal_gradient(self) -> float:
        """Material thermal gradient (K/m).

        Valid values are from the :obj:`MIN_THERMAL_GRADIENT` value to the
        :obj:`MAX_THERMAL_GRADIENT` value.
        """
        return self._thermal_gradient

    @thermal_gradient.setter
    def thermal_gradient(self, value: float):
        self.__validate_range(
            value,
            self.MIN_THERMAL_GRADIENT,
            self.MAX_THERMAL_GRADIENT,
            "thermal_gradient",
        )
        self._thermal_gradient = value

    @property
    def melt_pool_width(self) -> float:
        """Melt pool width (m).

        This is the width of the melt pool measured at the top of the powder layer.

        Valid values are from the :obj:`MIN_MELT_POOL_WIDTH` value to the
        :obj:`MAX_MELT_POOL_WIDTH` value.
        """
        return self._melt_pool_width

    @melt_pool_width.setter
    def melt_pool_width(self, value: float):
        self.__validate_range(
            value, self.MIN_MELT_POOL_WIDTH, self.MAX_MELT_POOL_WIDTH, "melt_pool_width"
        )
        self._melt_pool_width = value

    @property
    def melt_pool_depth(self) -> float:
        """Melt pool depth (m).

        This is the depth of the melt pool as measured from the top of the powder layer.

        Valid values are from the :obj:`MIN_MELT_POOL_DEPTH` value to the
        :obj:`MAX_MELT_POOL_DEPTH` value.
        """
        return self._melt_pool_depth

    @melt_pool_depth.setter
    def melt_pool_depth(self, value: float):
        self.__validate_range(
            value, self.MIN_MELT_POOL_DEPTH, self.MAX_MELT_POOL_DEPTH, "melt_pool_depth"
        )
        self._melt_pool_depth = value

    @property
    def random_seed(self) -> int:
        """Random seed for the simulation.

        Valid values are from the :obj:`MIN_RANDOM_SEED` value to the
        :obj:`MAX_RANDOM_SEED` value.
        """
        return self._random_seed

    @random_seed.setter
    def random_seed(self, value: int):
        self.__validate_range(value, self.MIN_RANDOM_SEED, self.MAX_RANDOM_SEED, "random_seed")
        self._random_seed = value

    def _to_simulation_request(self) -> SimulationRequest:
        """Convert this object into a simulation request message."""
        input = MicrostructureInputMessage(
            machine=self.machine._to_machine_message(),
            material=self.material._to_material_message(),
            cube_min_x=self.sample_min_x,
            cube_min_y=self.sample_min_y,
            cube_min_z=self.sample_min_z,
            cube_size_x=self.sample_size_x,
            cube_size_y=self.sample_size_y,
            cube_size_z=self.sample_size_z,
            sensor_dimension=self.sensor_dimension,
            use_provided_thermal_parameters=self.use_provided_thermal_parameters,
            cooling_rate=self.cooling_rate,
            thermal_gradient=self.thermal_gradient,
            melt_pool_width=self.melt_pool_width,
            melt_pool_depth=self.melt_pool_depth,
            use_random_seed=(self.random_seed != self.DEFAULT_RANDOM_SEED),
            random_seed=self.random_seed,
        )
        return SimulationRequest(id=self.id, microstructure_input=input)


class CircleEquivalenceColumnNames:
    """Provides column names for the circle equivalence data frame."""

    GRAIN_NUMBER = "grain_number"
    """Grain number."""
    AREA_FRACTION = "area_fraction"
    """Area fraction for grain."""
    DIAMETER = "diameter_um"
    """Grain diameter (µm)."""
    ORIENTATION_ANGLE = "orientation_angle"
    """Orientation angle (degrees)."""


class MicrostructureSummary(SimulationSummaryBase):
    """Provides the summary of a microstructure simulation."""

    def __init__(
        self,
        input: MicrostructureInput,
        result: MicrostructureResultMessage,
        logs: str,
        user_data_path: str,
        status: SimulationStatus = SimulationStatus.COMPLETED,
    ) -> None:
        """Initialize a ``MicrostructureSummary`` object."""
        if not isinstance(input, MicrostructureInput):
            raise ValueError("Invalid input type passed to init, " + self.__class__.__name__)
        if not isinstance(result, MicrostructureResultMessage):
            raise ValueError("Invalid result type passed to init, " + self.__class__.__name__)
        if not user_data_path:
            raise ValueError("Invalid user data path, " + self.__class__.__name__)
        if not isinstance(logs, str):
            raise ValueError("Invalid logs type passed to init, " + self.__class__.__name__)
        super().__init__(logs, status)
        self._input = input
        id = input.id if input.id else misc.short_uuid()
        outpath = os.path.join(user_data_path, id)
        self._result = Microstructure2DResult(result, outpath)
        self._logs = logs

    def __repr__(self):
        repr = type(self).__name__ + "\n"
        for k in [x for x in self.__dict__ if x != "_result"]:
            repr += k.replace("_", "", 1) + ": " + str(getattr(self, k)) + "\n"
        repr += self._result.__repr__()
        return repr

    @property
    def input(self):
        """Simulation input.

        For more information, see the :class:`MicrostructureInput` class.
        """
        return self._input

    @property
    def xy_vtk(self) -> str:
        """Path to the VTK file containing the 2D grain structure data in the XY plane.

        The VTK file contains these scalar data sets: ``GrainBoundaries``, ``Orientation_(deg)``, and
        ``GrainNumber``.
        """
        return self._result._xy_vtk

    @property
    def xz_vtk(self) -> str:
        """Path to the VTK file containing the 2D grain structure data in the XZ plane.

        The VTK file contains these scalar data sets: ``GrainBoundaries``,
        ``Orientation_(deg)``, and ``GrainNumber``.
        """
        return self._result._xz_vtk

    @property
    def yz_vtk(self) -> str:
        """Path to the VTK file containing the 2D grain structure data in the YZ plane.

        The VTK file contains these scalar data sets: ``GrainBoundaries``,
        ``Orientation_(deg)``, and ``GrainNumber``.
        """
        return self._result._yz_vtk

    @property
    def xy_circle_equivalence(self) -> pd.DataFrame:
        """Circle equivalence data for the XY plane.

        For data frame column names, see the :class:`CircleEquivalenceColumnNames` class.
        """
        return self._result._xy_circle_equivalence

    @property
    def xz_circle_equivalence(self) -> pd.DataFrame:
        """Circle equivalence data for the XZ plane.

        For data frame column names, see the :class:`CircleEquivalenceColumnNames` class.
        """
        return self._result._xz_circle_equivalence

    @property
    def yz_circle_equivalence(self) -> pd.DataFrame:
        """Circle equivalence data for the YZ plane.

        For data frame column names, see the :class:`CircleEquivalenceColumnNames` class.
        """
        return self._result._yz_circle_equivalence

    @property
    def xy_average_grain_size(self) -> float:
        """Average grain size (µm) for the XY plane."""
        return self._result._xy_average_grain_size

    @property
    def xz_average_grain_size(self) -> float:
        """Average grain size (µm) for the XZ plane."""
        return self._result._xz_average_grain_size

    @property
    def yz_average_grain_size(self) -> float:
        """Average grain size (µm) for the YZ plane."""
        return self._result._yz_average_grain_size

    @staticmethod
    def _circle_equivalence_frame(src: RepeatedCompositeFieldContainer) -> pd.DataFrame:
        d = {}
        d[CircleEquivalenceColumnNames.GRAIN_NUMBER] = np.asarray([x.grain_number for x in src])
        d[CircleEquivalenceColumnNames.AREA_FRACTION] = np.asarray([x.area_fraction for x in src])
        d[CircleEquivalenceColumnNames.DIAMETER] = np.asarray([x.diameter_um for x in src])
        d[CircleEquivalenceColumnNames.ORIENTATION_ANGLE] = np.asarray(
            [math.degrees(x.orientation_angle) for x in src]
        )
        return pd.DataFrame(d)

    @staticmethod
    def _average_grain_size(df: pd.DataFrame) -> float:
        """Calculate the average grain size (µm) for a given plane.

        Parameters
        ----------
        df : pd.DataFrame
            Data frame containing circle equivalence data for a given plane.

        Returns
        -------
        float
            Average grain size (µm).

        """

        return (
            df[CircleEquivalenceColumnNames.DIAMETER]
            * df[CircleEquivalenceColumnNames.AREA_FRACTION]
        ).sum()


class Microstructure2DResult:
    """Provides the results of a 2D microstructure simulation."""

    def __init__(self, msg: MicrostructureResultMessage, output_data_path: str) -> None:
        """Initialize a ``Microstructure2DResult`` object."""
        if not isinstance(msg, MicrostructureResultMessage):
            raise ValueError("Invalid msg parameter, " + self.__class__.__name__)
        if not output_data_path:
            raise ValueError("Invalid output data path, " + self.__class__.__name__)

        if not os.path.exists(output_data_path):
            os.makedirs(output_data_path)
        self._xy_vtk = os.path.join(output_data_path, "xy.vtk")
        with open(self._xy_vtk, "wb") as xy_vtk:
            xy_vtk.write(msg.xy_vtk)
        self._xz_vtk = os.path.join(output_data_path, "xz.vtk")
        with open(self._xz_vtk, "wb") as xz_vtk:
            xz_vtk.write(msg.xz_vtk)
        self._yz_vtk = os.path.join(output_data_path, "yz.vtk")
        with open(self._yz_vtk, "wb") as yz_vtk:
            yz_vtk.write(msg.yz_vtk)

        self._xy_circle_equivalence = MicrostructureSummary._circle_equivalence_frame(
            msg.xy_circle_equivalence
        )
        self._xz_circle_equivalence = MicrostructureSummary._circle_equivalence_frame(
            msg.xz_circle_equivalence
        )
        self._yz_circle_equivalence = MicrostructureSummary._circle_equivalence_frame(
            msg.yz_circle_equivalence
        )
        self._xy_average_grain_size = MicrostructureSummary._average_grain_size(
            self._xy_circle_equivalence
        )
        self._xz_average_grain_size = MicrostructureSummary._average_grain_size(
            self._xz_circle_equivalence
        )
        self._yz_average_grain_size = MicrostructureSummary._average_grain_size(
            self._yz_circle_equivalence
        )

    @property
    def xy_average_grain_size(self) -> float:
        """Average grain size (µm) for the XY plane."""
        return self._xy_average_grain_size

    @property
    def xz_average_grain_size(self) -> float:
        """Average grain size (µm) for the XZ plane."""
        return self._xz_average_grain_size

    @property
    def yz_average_grain_size(self) -> float:
        """Average grain size (µm) for the YZ plane."""
        return self._yz_average_grain_size

    def __repr__(self):
        repr = ""
        for k in self.__dict__:
            repr += k.replace("_", "", 1) + ": " + str(getattr(self, k)) + "\n"
        return repr



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides progress updates."""

from abc import ABC, abstractmethod
from enum import IntEnum
from os import getenv

from pydantic import BaseModel
from tqdm import tqdm

from ansys.api.additive.v0.additive_domain_pb2 import Progress as ProgressMsg
from ansys.api.additive.v0.additive_domain_pb2 import ProgressState as ProgressMsgState
from ansys.api.additive.v0.additive_operations_pb2 import OperationMetadata


class ProgressState(IntEnum):
    """Simulation progress status."""

    NEW = ProgressMsgState.PROGRESS_STATE_NEW
    """Simulation created and not yet queued to run."""
    WAITING = ProgressMsgState.PROGRESS_STATE_WAITING
    """Simulation is queued and waiting to start."""
    RUNNING = ProgressMsgState.PROGRESS_STATE_EXECUTING
    """Simulation is running."""
    COMPLETED = ProgressMsgState.PROGRESS_STATE_COMPLETED
    """Simulation has completed."""
    ERROR = ProgressMsgState.PROGRESS_STATE_ERROR
    """Simulation has errored."""
    CANCELLED = ProgressMsgState.PROGRESS_STATE_CANCELLED
    """Simulation has been cancelled."""
    WARNING = ProgressMsgState.PROGRESS_STATE_WARNING
    """Simulation completed with warnings."""


class Progress(BaseModel):
    """Progress information."""

    sim_id: str
    state: ProgressState
    percent_complete: int
    message: str
    context: str

    @classmethod
    def from_proto_msg(cls, sim_id: str, progress: ProgressMsg):
        """Create a ``Progress`` object from a progress protobuf message."""
        return cls(
            sim_id=sim_id,
            state=ProgressState(progress.state),
            percent_complete=progress.percent_complete,
            message=progress.message,
            context=progress.context,
        )

    @classmethod
    def from_operation_metadata(cls, metadata: OperationMetadata):
        """Create a ``Progress`` object from an operation metadata (long-running operations) protobuf message."""  # noqa: E501
        return cls(
            sim_id=metadata.simulation_id,
            state=metadata.state,
            percent_complete=metadata.percent_complete,
            message=metadata.message,
            context=metadata.context,
        )

    def __str__(self):
        return f"{self.sim_id}: {self.state.name} - {self.percent_complete}% - {self.context} - {self.message}"  # noqa: E501


class IProgressHandler(ABC):
    """Interface for simulation progress updates."""

    @abstractmethod
    def update(self, progress: Progress):
        """Update progress.

        Parameters
        ----------
        progress : Progress
            Progress information.

        """
        raise NotImplementedError


class DefaultSingleSimulationProgressHandler(IProgressHandler):
    """Creates a progress bar for a single simulation.

    Parameters
    ----------
    sim_id : str
        Simulation ID.

    """

    def __init__(self):
        """Initialize progress handler."""
        self._last_percent_complete = 0
        self._last_context = "Initializing"

    def update(self, progress: Progress):
        """Update the progress bar.

        Parameters
        ----------
        progress: Progress
            Latest progress.

        """
        # Don't send  progress when generating docs
        if getenv("GENERATING_DOCS"):
            return

        # Skip SOLVERINFO messages
        if progress.message and "SOLVERINFO" in progress.message:
            return

        if not hasattr(self, "_pbar"):
            self._pbar = tqdm(
                total=100,
                colour="green",
                desc=self._last_context,
                dynamic_ncols=True,
            )

        if progress.state == ProgressState.ERROR:
            self._pbar.write(progress.message)
            return

        if progress.context and progress.context != self._last_context:
            if "Solving Layer" not in progress.context or progress.context == "Solving Layer 1":
                self._pbar.reset(total=100)
                self._pbar.set_description(progress.context)
                self._last_context = progress.context
                self._last_percent_complete = 0
            else:
                self._pbar.set_description(progress.context, refresh=False)

        if progress.percent_complete - self._last_percent_complete > 0:
            self._pbar.update(progress.percent_complete - self._last_percent_complete)
        self._last_percent_complete = progress.percent_complete

    def __del__(self):
        if hasattr(self, "_pbar"):
            self._pbar.close()



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Manages simulation tasks."""

import time

from ansys.additive.core.logger import LOG
from ansys.additive.core.progress_handler import IProgressHandler, Progress
from ansys.additive.core.simulation_task import SimulationTask


class SimulationTaskManager:
    """Provides a manager for simulation tasks."""

    def __init__(self):
        """Initialize the simulation task manager."""
        self._tasks: list[SimulationTask] = []

    @property
    def tasks(self) -> list[SimulationTask]:
        """Get the list of tasks managed by this manager."""
        return self._tasks

    @property
    def simulation_ids(self) -> list[str]:
        """Get the list of the simulation ids managed by this manager."""
        return [x.simulation_id for x in self._tasks]

    @property
    def done(self) -> bool:
        """Check if all tasks are done."""
        return all(t.done for t in self._tasks)

    def add_task(self, task: SimulationTask):
        """Add a task to this manager.

        Parameters
        ----------
        task: SimulationTask
            The simulation task holding the long-running operation and corresponding server.

        """
        self._tasks.append(task)

    def status(
        self, progress_handler: IProgressHandler | None = None
    ) -> list[tuple[str, Progress]]:
        """Get status of each operation stored in this manager.

        Parameters
        ----------
        progress_handler: IProgressHandler, None, default: None
            Handler for progress updates. If ``None``, no progress updates are provided.

        Returns
        -------
        List of tuples with each tuple containing the operation name and an instance of Progress

        """
        status_all = []

        for t in self._tasks:
            progress = t.status()
            status_all.append((progress.sim_id, progress))
            if progress_handler:
                progress_handler.update(progress)

        return status_all

    def wait_all(self, progress_handler: IProgressHandler | None = None) -> None:
        """Wait for all simulations to finish. A simple loop that waits for each task will wait for the
        simulation that takes the longest. This works because wait returns immediately if operation is done.

        Parameters
        ----------
        progress_handler: IProgressHandler, None, default: None
            Handler for progress updates. If ``None``, no progress updates are provided.

        """
        LOG.debug(f"Waiting for {len(self._tasks)} tasks to complete")
        for t in self._tasks:
            t.wait(progress_handler=progress_handler)

    def cancel_all(self) -> None:
        """Cancel all simulations belonging to this simulation task manager."""
        LOG.debug("Cancelling all tasks")
        for t in self._tasks:
            t.cancel()
            time.sleep(0.1)

    def summaries(self):
        """Get a list of the summaries of completed simulations only."""
        return [t.summary for t in self._tasks if t.summary]



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides a container for material parameters."""

import collections
import csv
import json
import re

from ansys.api.additive.v0.additive_domain_pb2 import (
    AdditiveMaterial as MaterialMessage,
)
from ansys.api.additive.v0.additive_domain_pb2 import (
    CharacteristicWidthDataPoint as CharacteristicWidthDataPointMessage,
)
from ansys.api.additive.v0.additive_domain_pb2 import (
    ThermalPropertiesDataPoint as ThermalPropertiesDataPointMessage,
)

RESERVED_MATERIAL_NAMES = [
    "17-4PH",
    "316L",
    "Al357",
    "AlSi10Mg",
    "CoCr",
    "IN625",
    "IN718",
    "Ti64",
]


class CharacteristicWidthDataPoint:
    """Provides the container for a characteristic width data point.

    Additive material definitions include a file containing a characteristic width
    lookup table, allowing a given laser speed and power to be correlated to a
    characteristic melt pool width. This class represents a single row in the lookup
    table.
    """

    def __init__(
        self,
        *,
        laser_power: float = 0,
        scan_speed: float = 0,
        characteristic_width: float = 0,
    ):
        """Initialize a characteristic width data point."""
        self._laser_power = laser_power
        self._scan_speed = scan_speed
        self._characteristic_width = characteristic_width

    def __repr__(self):
        repr = type(self).__name__ + "\n"
        for k in self.__dict__:
            repr += k.replace("_", "", 1) + ": " + str(getattr(self, k)) + "\n"
        return repr

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CharacteristicWidthDataPoint):
            return False
        return all(getattr(self, k) == getattr(other, k) for k in self.__dict__)

    @property
    def characteristic_width(self) -> float:
        """Characteristic melt pool width for a given laser power and scan
        speed (m).
        """
        return self._characteristic_width

    @characteristic_width.setter
    def characteristic_width(self, value: float):
        """Set characteristic width value."""
        if value < 0:
            raise ValueError("Characteristic width must not be negative.")
        self._characteristic_width = value

    @property
    def laser_power(self) -> float:
        """Laser power (W)."""
        return self._laser_power

    @laser_power.setter
    def laser_power(self, value: float):
        """Set power value."""
        if value < 0:
            raise ValueError("Power must not be negative.")
        self._laser_power = value

    @property
    def scan_speed(self) -> float:
        """Laser scan speed (m/s)."""
        return self._scan_speed

    @scan_speed.setter
    def scan_speed(self, value: float):
        """Set speed value."""
        if value < 0:
            raise ValueError("Speed must not be negative.")
        self._scan_speed = value

    @staticmethod
    def _from_characteristic_width_data_point_message(
        msg: CharacteristicWidthDataPointMessage,
    ):
        """Create a characteristic width data point`` from a characteristic
        data point message received from the Additive service.
        """
        if not isinstance(msg, CharacteristicWidthDataPointMessage):
            raise ValueError(
                "Invalid message object passed to from_characteristic_width_data_point_message()"
            )
        point = CharacteristicWidthDataPoint()
        for p in point.__dict__:
            setattr(point, p, getattr(msg, p.replace("_", "", 1)))
        return point

    def _to_characteristic_width_data_point_message(
        self,
    ) -> CharacteristicWidthDataPointMessage:
        """Create a characteristic width data point message from this
        characteristic width data point to send to the Additive service.
        """
        msg = CharacteristicWidthDataPointMessage()
        for p in self.__dict__:
            setattr(msg, p.replace("_", "", 1), getattr(self, p))
        return msg


class ThermalPropertiesDataPoint:
    """Provides the container for temperature-dependent properties.

    Additive material definitions include a file containing a lookup table describing
    the material's thermal properties at different temperatures. This class represents a
    single row in the lookup table.

    Units are SI (m, kg, s, K) unless otherwise noted.
    """

    def __init__(
        self,
        *,
        density: float = 0,
        density_ratio: float = 0,
        specific_heat: float = 0,
        specific_heat_ratio: float = 0,
        temperature: float = 0,
        thermal_conductivity: float = 0,
        thermal_conductivity_ratio: float = 0,
    ):
        """Create a thermal properties data point."""
        self._density = density
        self._density_ratio = density_ratio
        self._specific_heat = specific_heat
        self._specific_heat_ratio = specific_heat_ratio
        self._temperature = temperature
        self._thermal_conductivity = thermal_conductivity
        self._thermal_conductivity_ratio = thermal_conductivity_ratio

    def __repr__(self):
        repr = type(self).__name__ + "\n"
        for k in self.__dict__:
            repr += k.replace("_", "", 1) + ": " + str(getattr(self, k)) + "\n"
        return repr

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ThermalPropertiesDataPoint):
            return False
        return all(getattr(self, k) == getattr(other, k) for k in self.__dict__)

    @property
    def density(self) -> float:
        """Density (kg/m^3)."""
        return self._density

    @density.setter
    def density(self, value: float):
        """Set density value."""
        if value < 0:
            raise ValueError("Density must not be negative.")
        self._density = value

    @property
    def density_ratio(self) -> float:
        """Density ratio."""
        return self._density_ratio

    @density_ratio.setter
    def density_ratio(self, value: float):
        """Set density ratio value."""
        if value < 0:
            raise ValueError("Density ratio must not be negative.")
        self._density_ratio = value

    @property
    def specific_heat(self) -> float:
        """Specific heat (J/kg/K)."""
        return self._specific_heat

    @specific_heat.setter
    def specific_heat(self, value: float):
        """Set specific heat."""
        self._specific_heat = value

    @property
    def specific_heat_ratio(self) -> float:
        """Specific heat ratio."""
        return self._specific_heat_ratio

    @specific_heat_ratio.setter
    def specific_heat_ratio(self, value: float):
        """Set specific heat ratio."""
        self._specific_heat_ratio = value

    @property
    def temperature(self) -> float:
        """Temperature (K)."""
        return self._temperature

    @temperature.setter
    def temperature(self, value: float):
        """Set temperature."""
        if value < 0:
            raise ValueError("Temperature must not be negative.")
        self._temperature = value

    @property
    def thermal_conductivity(self) -> float:
        """Thermal conductivity (W/m/K)."""
        return self._thermal_conductivity

    @thermal_conductivity.setter
    def thermal_conductivity(self, value: float):
        """Set thermal conductivity."""
        self._thermal_conductivity = value

    @property
    def thermal_conductivity_ratio(self) -> float:
        """Thermal conductivity ratio."""
        return self._thermal_conductivity_ratio

    @thermal_conductivity_ratio.setter
    def thermal_conductivity_ratio(self, value: float):
        """Set thermal conductivity ratio."""
        self._thermal_conductivity_ratio = value

    @staticmethod
    def _from_thermal_properties_data_point_message(
        msg: ThermalPropertiesDataPointMessage,
    ):
        """Create a thermal properties data point from a thermal characteristic
        data point message received from the Additive service.

        :meta private:
        """
        if not isinstance(msg, ThermalPropertiesDataPointMessage):
            raise ValueError(
                "Invalid message object passed to from_thermal_properties_data_point_message()"
            )
        point = ThermalPropertiesDataPoint()
        for p in point.__dict__:
            setattr(point, p, getattr(msg, p.replace("_", "", 1)))
        return point

    def _to_thermal_properties_data_point_message(
        self,
    ) -> ThermalPropertiesDataPointMessage:
        """Create a thermal characteristic data point message from this thermal
        properties data point to send to the Additive service.

        :meta private:
        """
        msg = ThermalPropertiesDataPointMessage()
        for p in self.__dict__:
            setattr(msg, p.replace("_", "", 1), getattr(self, p))
        return msg


class AdditiveMaterial:
    """Provides the container for material properties used during additive manufacturing simulations."""

    def __init__(
        self,
        *,
        absorptivity_maximum: float = 0,
        absorptivity_minimum: float = 0,
        absorptivity_powder_coefficient_a: float = 0,
        absorptivity_powder_coefficient_b: float = 0,
        absorptivity_solid_coefficient_a: float = 0,
        absorptivity_solid_coefficient_b: float = 0,
        anisotropic_strain_coefficient_parallel: float = 0,
        anisotropic_strain_coefficient_perpendicular: float = 0,
        anisotropic_strain_coefficient_z: float = 0,
        description: str = "",
        elastic_modulus: float = 0,
        hardening_factor: float = 0,
        liquidus_temperature: float = 0,
        material_yield_strength: float = 0,
        name: str = "",
        nucleation_constant_bulk: float = 0,
        nucleation_constant_interface: float = 0,
        penetration_depth_maximum: float = 0,
        penetration_depth_minimum: float = 0,
        penetration_depth_powder_coefficient_a: float = 0,
        penetration_depth_powder_coefficient_b: float = 0,
        penetration_depth_solid_coefficient_a: float = 0,
        penetration_depth_solid_coefficient_b: float = 0,
        poisson_ratio: float = 0,
        powder_packing_density: float = 0,
        purging_gas_convection_coefficient: float = 0,
        solid_density_at_room_temperature: float = 0,
        solid_specific_heat_at_room_temperature: float = 0,
        solid_thermal_conductivity_at_room_temperature: float = 0,
        solidus_temperature: float = 0,
        strain_scaling_factor: float = 0,
        support_yield_strength_ratio: float = 0,
        thermal_expansion_coefficient: float = 0,
        vaporization_temperature: float = 0,
        characteristic_width_data: list[CharacteristicWidthDataPoint] = None,
        thermal_properties_data: list[ThermalPropertiesDataPoint] = None,
    ):
        """Create an additive material."""
        self._absorptivity_maximum = absorptivity_maximum
        self._absorptivity_minimum = absorptivity_minimum
        self._absorptivity_powder_coefficient_a = absorptivity_powder_coefficient_a
        self._absorptivity_powder_coefficient_b = absorptivity_powder_coefficient_b
        self._absorptivity_solid_coefficient_a = absorptivity_solid_coefficient_a
        self._absorptivity_solid_coefficient_b = absorptivity_solid_coefficient_b
        self._anisotropic_strain_coefficient_parallel = anisotropic_strain_coefficient_parallel
        self._anisotropic_strain_coefficient_perpendicular = (
            anisotropic_strain_coefficient_perpendicular
        )
        self._anisotropic_strain_coefficient_z = anisotropic_strain_coefficient_z
        self._description = description
        self._elastic_modulus = elastic_modulus
        self._hardening_factor = hardening_factor
        self._liquidus_temperature = liquidus_temperature
        self._material_yield_strength = material_yield_strength
        self._name = name
        self._nucleation_constant_bulk = nucleation_constant_bulk
        self._nucleation_constant_interface = nucleation_constant_interface
        self._penetration_depth_maximum = penetration_depth_maximum
        self._penetration_depth_minimum = penetration_depth_minimum
        self._penetration_depth_powder_coefficient_a = penetration_depth_powder_coefficient_a
        self._penetration_depth_powder_coefficient_b = penetration_depth_powder_coefficient_b
        self._penetration_depth_solid_coefficient_a = penetration_depth_solid_coefficient_a
        self._penetration_depth_solid_coefficient_b = penetration_depth_solid_coefficient_b
        self._poisson_ratio = poisson_ratio
        self._powder_packing_density = powder_packing_density
        self._purging_gas_convection_coefficient = purging_gas_convection_coefficient
        self._solid_density_at_room_temperature = solid_density_at_room_temperature
        self._solid_specific_heat_at_room_temperature = solid_specific_heat_at_room_temperature
        self._solid_thermal_conductivity_at_room_temperature = (
            solid_thermal_conductivity_at_room_temperature
        )
        self._solidus_temperature = solidus_temperature
        self._strain_scaling_factor = strain_scaling_factor
        self._support_yield_strength_ratio = support_yield_strength_ratio
        self._thermal_expansion_coefficient = thermal_expansion_coefficient
        self._vaporization_temperature = vaporization_temperature
        if characteristic_width_data:
            self._characteristic_width_data = characteristic_width_data
        else:
            self._characteristic_width_data = []
        if thermal_properties_data:
            self._thermal_properties_data = thermal_properties_data
        else:
            self._thermal_properties_data = []

    def __repr__(self) -> str:
        repr = self.__class__.__name__ + "\n"
        for p in self.__dict__:
            if p != "_characteristic_width_data" and p != "_thermal_properties_data":
                repr = repr + "{}: {}\n".format(p.replace("_", "", 1), getattr(self, p))
        repr = (
            repr
            + "characteristic_width_data: CharacteristicWidthDataPoint[]\n"
            + "thermal_properties_data: ThermalPropertiesDataPoint[]\n"
        )
        return repr

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, AdditiveMaterial):
            return False
        return all(getattr(self, k) == getattr(__o, k) for k in self.__dict__)

    @property
    def absorptivity_maximum(self) -> float:
        """Absorptivity maximum."""
        return self._absorptivity_maximum

    @absorptivity_maximum.setter
    def absorptivity_maximum(self, value: float):
        """Set absorptivity maximum."""
        self._absorptivity_maximum = value

    @property
    def absorptivity_minimum(self) -> float:
        """Absorptivity minimum."""
        return self._absorptivity_minimum

    @absorptivity_minimum.setter
    def absorptivity_minimum(self, value: float):
        """Set absorptivity minimum."""
        self._absorptivity_minimum = value

    @property
    def absorptivity_powder_coefficient_a(self) -> float:
        """Absorptivity powder coefficient a."""
        return self._absorptivity_powder_coefficient_a

    @absorptivity_powder_coefficient_a.setter
    def absorptivity_powder_coefficient_a(self, value: float):
        """Set absorptivity powder coefficient a."""
        self._absorptivity_powder_coefficient_a = value

    @property
    def absorptivity_powder_coefficient_b(self) -> float:
        """Absorptivity powder coefficient b."""
        return self._absorptivity_powder_coefficient_b

    @absorptivity_powder_coefficient_b.setter
    def absorptivity_powder_coefficient_b(self, value: float):
        """Set absorptivity powder coefficient b."""
        self._absorptivity_powder_coefficient_b = value

    @property
    def absorptivity_solid_coefficient_a(self) -> float:
        """Absorptivity solid coefficient a."""
        return self._absorptivity_solid_coefficient_a

    @absorptivity_solid_coefficient_a.setter
    def absorptivity_solid_coefficient_a(self, value: float):
        """Set absorptivity solid coefficient a."""
        self._absorptivity_solid_coefficient_a = value

    @property
    def absorptivity_solid_coefficient_b(self) -> float:
        """Absorptivity solid coefficient b."""
        return self._absorptivity_solid_coefficient_b

    @absorptivity_solid_coefficient_b.setter
    def absorptivity_solid_coefficient_b(self, value: float):
        """Set absorptivity solid coefficient b."""
        self._absorptivity_solid_coefficient_b = value

    @property
    def anisotropic_strain_coefficient_parallel(self) -> float:
        """Multiplier on the predicted strain in the direction that the laser is scanning for the major fill rasters."""  # noqa: E501
        return self._anisotropic_strain_coefficient_parallel

    @anisotropic_strain_coefficient_parallel.setter
    def anisotropic_strain_coefficient_parallel(self, value: float):
        """Set anisotropic strain coefficient parallel."""
        self._anisotropic_strain_coefficient_parallel = value

    @property
    def anisotropic_strain_coefficient_perpendicular(self) -> float:
        """Multiplier on the predicted strain orthogonal to the direction that the laser is scanning for the major fill rasters and in the plane of the surface of the build plate."""  # noqa: E501
        return self._anisotropic_strain_coefficient_perpendicular

    @anisotropic_strain_coefficient_perpendicular.setter
    def anisotropic_strain_coefficient_perpendicular(self, value: float):
        """Set anisotropic strain coefficient perpendicular."""
        self._anisotropic_strain_coefficient_perpendicular = value

    @property
    def anisotropic_strain_coefficient_z(self) -> float:
        """Multiplier on the predicted strain in the Z direction."""
        return self._anisotropic_strain_coefficient_z

    @anisotropic_strain_coefficient_z.setter
    def anisotropic_strain_coefficient_z(self, value: float):
        """Set anisotropic strain coefficient z."""
        self._anisotropic_strain_coefficient_z = value

    @property
    def description(self) -> str:
        """Description of the material."""
        return self._description

    @description.setter
    def description(self, value: str):
        """Set description."""
        self._description = value

    @property
    def elastic_modulus(self) -> float:
        """Elastic modulus (Pa)."""
        return self._elastic_modulus

    @elastic_modulus.setter
    def elastic_modulus(self, value: float):
        """Set elastic modulus."""
        self._elastic_modulus = value

    @property
    def hardening_factor(self) -> float:
        """Factor relating the elastic modulus to the tangent modulus for plasticity simulations (tangent modulus = elastic modulus * hardening factor)."""  # noqa: E501
        return self._hardening_factor

    @hardening_factor.setter
    def hardening_factor(self, value: float):
        """Set hardening factor."""
        self._hardening_factor = value

    @property
    def liquidus_temperature(self) -> float:
        """Minimum temperature (K) at which the material is completely liquid."""
        return self._liquidus_temperature

    @liquidus_temperature.setter
    def liquidus_temperature(self, value: float):
        """Set liquidus temperature."""
        self._liquidus_temperature = value

    @property
    def material_yield_strength(self) -> float:
        """Material yield strength (Pa)."""
        return self._material_yield_strength

    @material_yield_strength.setter
    def material_yield_strength(self, value: float):
        """Set material yield strength."""
        self._material_yield_strength = value

    @property
    def name(self) -> str:
        """Name of the material."""
        return self._name

    @name.setter
    def name(self, value: str):
        """Set name."""
        self._name = value

    @property
    def nucleation_constant_bulk(self) -> float:
        """Controls the homogeneous nucleation rate (in bulk of the microstructure simulation domain) during solidification (1/m^2/K^2)."""  # noqa: E501
        return self._nucleation_constant_bulk

    @nucleation_constant_bulk.setter
    def nucleation_constant_bulk(self, value: float):
        """Set nucleation constant bulk."""
        self._nucleation_constant_bulk = value

    @property
    def nucleation_constant_interface(self) -> float:
        """Heterogeneous nucleation rate (on existing solid interfaces) during solidification (1/m/K^2)."""  # noqa: E501
        return self._nucleation_constant_interface

    @nucleation_constant_interface.setter
    def nucleation_constant_interface(self, value: float):
        """Set nucleation constant interface."""
        self._nucleation_constant_interface = value

    @property
    def penetration_depth_maximum(self) -> float:
        """Penetration depth maximum."""
        return self._penetration_depth_maximum

    @penetration_depth_maximum.setter
    def penetration_depth_maximum(self, value: float):
        """Set penetration depth maximum."""
        self._penetration_depth_maximum = value

    @property
    def penetration_depth_minimum(self) -> float:
        """Penetration depth minimum."""
        return self._penetration_depth_minimum

    @penetration_depth_minimum.setter
    def penetration_depth_minimum(self, value: float):
        """Set penetration depth minimum."""
        self._penetration_depth_minimum = value

    @property
    def penetration_depth_powder_coefficient_a(self) -> float:
        """Penetration depth powder coefficient a."""
        return self._penetration_depth_powder_coefficient_a

    @penetration_depth_powder_coefficient_a.setter
    def penetration_depth_powder_coefficient_a(self, value: float):
        """Set penetration depth powder coefficient a."""
        self._penetration_depth_powder_coefficient_a = value

    @property
    def penetration_depth_powder_coefficient_b(self) -> float:
        """Penetration depth powder coefficient b."""
        return self._penetration_depth_powder_coefficient_b

    @penetration_depth_powder_coefficient_b.setter
    def penetration_depth_powder_coefficient_b(self, value: float):
        """Set penetration depth powder coefficient b."""
        self._penetration_depth_powder_coefficient_b = value

    @property
    def penetration_depth_solid_coefficient_a(self) -> float:
        """Penetration depth solid coefficient a."""
        return self._penetration_depth_solid_coefficient_a

    @penetration_depth_solid_coefficient_a.setter
    def penetration_depth_solid_coefficient_a(self, value: float):
        """Set penetration depth solid coefficient a."""
        self._penetration_depth_solid_coefficient_a = value

    @property
    def penetration_depth_solid_coefficient_b(self) -> float:
        """Penetration depth solid coefficient b."""
        return self._penetration_depth_solid_coefficient_b

    @penetration_depth_solid_coefficient_b.setter
    def penetration_depth_solid_coefficient_b(self, value: float):
        """Set penetration_depth solid coefficient b."""
        self._penetration_depth_solid_coefficient_b = value

    @property
    def poisson_ratio(self) -> float:
        """Poisson ratio."""
        return self._poisson_ratio

    @poisson_ratio.setter
    def poisson_ratio(self, value: float):
        """Set Poisson ratio."""
        self._poisson_ratio = value

    @property
    def powder_packing_density(self) -> float:
        """Density of powder material relative to the solid."""
        return self._powder_packing_density

    @powder_packing_density.setter
    def powder_packing_density(self, value: float):
        """Set powder packing density."""
        self._powder_packing_density = value

    @property
    def purging_gas_convection_coefficient(self) -> float:
        """Convection coefficient between the solid and gas during build."""
        return self._purging_gas_convection_coefficient

    @purging_gas_convection_coefficient.setter
    def purging_gas_convection_coefficient(self, value: float):
        """Set purging gas convection coefficient."""
        self._purging_gas_convection_coefficient = value

    @property
    def solid_density_at_room_temperature(self) -> float:
        """Density of bulk material at room temperature, 298 K (kg/m^3)."""
        return self._solid_density_at_room_temperature

    @solid_density_at_room_temperature.setter
    def solid_density_at_room_temperature(self, value: float):
        """Set solid density at room temperature."""
        self._solid_density_at_room_temperature = value

    @property
    def solid_specific_heat_at_room_temperature(self) -> float:
        """Specific heat of bulk material at room temperature, 298 K (J/kg/K)."""
        return self._solid_specific_heat_at_room_temperature

    @solid_specific_heat_at_room_temperature.setter
    def solid_specific_heat_at_room_temperature(self, value: float):
        """Set solid specific heat at room temperature."""
        self._solid_specific_heat_at_room_temperature = value

    @property
    def solid_thermal_conductivity_at_room_temperature(self) -> float:
        """Thermal conductivity of bulk material at room temperature, 298 K (W/m/K)."""
        return self._solid_thermal_conductivity_at_room_temperature

    @solid_thermal_conductivity_at_room_temperature.setter
    def solid_thermal_conductivity_at_room_temperature(self, value: float):
        """Set solid thermal conductivity at_room temperature."""
        self._solid_thermal_conductivity_at_room_temperature = value

    @property
    def solidus_temperature(self) -> float:
        """Maximum temperature (K) at which the material is completely solid."""
        return self._solidus_temperature

    @solidus_temperature.setter
    def solidus_temperature(self, value: float):
        """Set solidus temperature."""
        self._solidus_temperature = value

    @property
    def strain_scaling_factor(self) -> float:
        """Strain scaling factor."""
        return self._strain_scaling_factor

    @strain_scaling_factor.setter
    def strain_scaling_factor(self, value: float):
        """Set strain scaling factor."""
        self._strain_scaling_factor = value

    @property
    def support_yield_strength_ratio(self) -> float:
        """Knockdown factor that is used to adjust the strength of the support material in comparison to the solid material.

        The factor is multiplied by the support material's yield
        strength and elastic modulus. A value of 1.0, for example,
        results in support strength equal to the solid material, whereas
        0.5 is half the strength of the solid material.
        """  # noqa: E501
        return self._support_yield_strength_ratio

    @support_yield_strength_ratio.setter
    def support_yield_strength_ratio(self, value: float):
        """Set support yield strength ratio."""
        self._support_yield_strength_ratio = value

    @property
    def thermal_expansion_coefficient(self) -> float:
        """Coefficient of thermal expansion (1/K)."""
        return self._thermal_expansion_coefficient

    @thermal_expansion_coefficient.setter
    def thermal_expansion_coefficient(self, value: float):
        """Set thermal expansion coefficient."""
        self._thermal_expansion_coefficient = value

    @property
    def vaporization_temperature(self) -> float:
        """Temperature (K) at which the material has completely changed from liquid to vapor."""
        return self._vaporization_temperature

    @vaporization_temperature.setter
    def vaporization_temperature(self, value: float):
        """Set vaporization temperature (K)."""
        self._vaporization_temperature = value

    @property
    def characteristic_width_data(self) -> list[CharacteristicWidthDataPoint]:
        """List of characteristic width data points."""
        return self._characteristic_width_data

    @characteristic_width_data.setter
    def characteristic_width_data(self, value: list[CharacteristicWidthDataPoint]):
        """Set characteristic width data."""
        if not isinstance(value, collections.abc.Sequence):
            raise TypeError(
                "Invalid object type, {}, passed to characteristic_width_data()".format(type(value))
            )
        self._characteristic_width_data = value

    @property
    def thermal_properties_data(self) -> list[ThermalPropertiesDataPoint]:
        """List of thermal properties data points."""
        return self._thermal_properties_data

    @thermal_properties_data.setter
    def thermal_properties_data(self, value: list[ThermalPropertiesDataPoint]):
        """Set thermal properties data."""
        if not isinstance(value, collections.abc.Sequence):
            raise TypeError(
                "Invalid object type, {}, passed to thermal_properties_data()".format(type(value))
            )
        self._thermal_properties_data = value

    @staticmethod
    def _from_material_message(msg: MaterialMessage):
        """Create an additive material from a material message received from
        the Additive service.
        """
        if not isinstance(msg, MaterialMessage):
            raise ValueError("Invalid message object passed to from_material_message()")
        material = AdditiveMaterial()
        for p in material.__dict__:
            if p != "_characteristic_width_data" and p != "_thermal_properties_data":
                setattr(material, p, getattr(msg, p.replace("_", "", 1)))
        for c in msg.characteristic_width_data_points:
            material.characteristic_width_data.append(
                CharacteristicWidthDataPoint._from_characteristic_width_data_point_message(c)
            )
        for t in msg.thermal_properties_data_points:
            material.thermal_properties_data.append(
                ThermalPropertiesDataPoint._from_thermal_properties_data_point_message(t)
            )
        return material

    def _to_material_message(self) -> MaterialMessage:
        """Create a material message from the additive material to send to the
        Additive service.
        """
        msg = MaterialMessage()
        for p in self.__dict__:
            if p != "_characteristic_width_data" and p != "_thermal_properties_data":
                setattr(msg, p.replace("_", "", 1), getattr(self, p))
        for c in self.characteristic_width_data:
            msg.characteristic_width_data_points.append(
                c._to_characteristic_width_data_point_message()
            )
        for t in self.thermal_properties_data:
            msg.thermal_properties_data_points.append(t._to_thermal_properties_data_point_message())
        return msg

    def _load_parameters(self, parameters_file: str):
        """Load material parameters from a JSON file."""
        with open(parameters_file, "r") as f:
            data = json.load(f)
        self.name = data["name"]
        self.description = data["description"]
        parameters = data["configuration"]
        # Convert camelCase to snake_case
        pattern = re.compile(r"(?<!^)(?=[A-Z])")
        for p in parameters:
            if p != "materialName" and p != "elasticModulusOfBase":
                name = pattern.sub("_", p).lower()
                name = name.replace("_coeff_", "_coefficient_")
                setattr(self, name, parameters[p])

    def _load_thermal_properties(self, thermal_lookup_file: str):
        """Load thermal properties from a CSV file."""
        with open(thermal_lookup_file, "r") as f:
            reader = csv.reader(f)
            self.thermal_properties_data.clear()
            next(reader)  # skip header
            for row in reader:
                self.thermal_properties_data.append(
                    ThermalPropertiesDataPoint(
                        temperature=float(row[0]),
                        thermal_conductivity=float(row[1]),
                        specific_heat=float(row[2]),
                        density=float(row[3]),
                        thermal_conductivity_ratio=float(row[4]),
                        density_ratio=float(row[5]),
                        specific_heat_ratio=float(row[6]),
                    )
                )

    def _load_characteristic_width(self, cw_lookup_file: str):
        """Load characteristic width values from a CSV file."""
        with open(cw_lookup_file, "r") as f:
            reader = csv.reader(f)
            self.characteristic_width_data.clear()
            next(reader)
            for row in reader:
                self.characteristic_width_data.append(
                    CharacteristicWidthDataPoint(
                        scan_speed=float(row[0]),
                        laser_power=float(row[1]),
                        characteristic_width=float(row[2]),
                    )
                )



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides common definitions and classes for simulations."""

from enum import Enum

from ansys.additive.core.misc import short_uuid


class SimulationType(str, Enum):
    """Provides simulation types."""

    SINGLE_BEAD = "SingleBead"
    """Single bead simulation."""
    POROSITY = "Porosity"
    """Porosity simulation."""
    MICROSTRUCTURE = "Microstructure"
    """Microstructure simulation."""


class SimulationStatus(str, Enum):
    """Provides simulation status values."""

    # NOTE: Values are listed in order of precedence when deduping.
    # For example, if duplicate COMPLETED and WARNING simulations are
    # found when removing duplicates, the COMPLETED simulation will be
    # kept.
    COMPLETED = "Completed"
    """Simulation completed successfully."""
    WARNING = "Warning"
    """Simulation completed with warnings."""
    ERROR = "Error"
    """Simulation errored before completion."""
    CANCELLED = "Cancelled"
    """Simulation was cancelled."""
    RUNNING = "Running"
    """Simulation is running."""
    PENDING = "Pending"
    """Simulation is queued and waiting to run."""
    NEW = "New"
    """Simulation is created but not yet queued to run."""
    SKIP = "Skip"
    """Do not run this simulation, only applies to parametric studies."""


class SimulationInputBase:
    """Provides a base class for simulation inputs."""

    def __init__(self) -> None:
        """Initialize the simulation input base class."""
        self._id: str = short_uuid()

    @property
    def id(self) -> str:
        """Return a unique identifier for this simulation."""
        return self._id


class SimulationSummaryBase:
    """Provides a base class for simulation summaries."""

    def __init__(self, logs: str, status: SimulationStatus = SimulationStatus.COMPLETED):
        """Initialize a ``SimulationSummaryBase`` object."""
        self._logs = logs
        self._status = status

    @property
    def logs(self) -> str:
        """Simulation logs."""
        return self._logs

    @property
    def status(self) -> SimulationStatus:
        """Simulation status."""
        return self._status



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Set up methods for grpc simulation requests."""

import hashlib
import os
from collections.abc import Iterator

from ansys.additive.core.microstructure import MicrostructureInput
from ansys.additive.core.microstructure_3d import Microstructure3DInput
from ansys.additive.core.porosity import PorosityInput
from ansys.additive.core.progress_handler import (
    IProgressHandler,
    Progress,
    ProgressState,
)
from ansys.additive.core.server_connection import ServerConnection
from ansys.additive.core.single_bead import SingleBeadInput
from ansys.additive.core.thermal_history import ThermalHistoryInput
from ansys.api.additive.v0.additive_simulation_pb2 import (
    SimulationRequest,
    UploadFileRequest,
)


def __file_upload_reader(file_name: str, chunk_size=2 * 1024**2) -> Iterator[UploadFileRequest]:
    """Read a file and return an iterator of UploadFileRequests."""
    file_size = os.path.getsize(file_name)
    short_name = os.path.basename(file_name)
    with open(file_name, mode="rb") as f:
        while True:
            chunk = f.read(chunk_size)
            if not chunk:
                break
            yield UploadFileRequest(
                name=short_name,
                total_size=file_size,
                content=chunk,
                content_md5=hashlib.md5(chunk).hexdigest(),  # noqa: S324
            )


def _setup_thermal_history(
    input: ThermalHistoryInput,
    server: ServerConnection,
    progress_handler: IProgressHandler | None = None,
) -> SimulationRequest:
    """Initialize a thermal history simulation.

    Parameters
    ----------
    input: ThermalHistoryInput
        Simulation input parameters.
    server: ServerConnection
        Server to use for the simulation.
    progress_handler: IProgressHandler, None, default: None
        Handler for progress updates. If ``None``, no progress updates are provided.

    Returns
    -------
    :class:`SimulationRequest`

    """
    if not input.geometry or not input.geometry.path:
        raise ValueError("The geometry path is not defined in the simulation input")

    remote_geometry_path = ""
    for response in server.simulation_stub.UploadFile(__file_upload_reader(input.geometry.path)):
        remote_geometry_path = response.remote_file_name
        progress = Progress.from_proto_msg(input.id, response.progress)
        if progress_handler:
            progress_handler.update(progress)
        if progress.state == ProgressState.ERROR:
            raise Exception(progress.message)

    return input._to_simulation_request(remote_geometry_path=remote_geometry_path)


def create_request(
    simulation_input: (
        SingleBeadInput
        | PorosityInput
        | MicrostructureInput
        | ThermalHistoryInput
        | Microstructure3DInput
    ),
    server: ServerConnection,
    progress_handler: IProgressHandler | None = None,
) -> SimulationRequest:
    """Create a simulation request and set up any pre-requisites on a server, such as an STL file for a
    thermal history simulation.

    Parameters
    ----------
    simulation_input: SingleBeadInput, PorosityInput, MicrostructureInput, ThermalHistoryInput, Microstructure3DInput
        Parameters to use for simulation.
    server: ServerConnection
        Server to use for the simulation.
    progress_handler: IProgressHandler, None, default: None
        Handler for progress updates. If ``None``, no progress updates are provided.

    Returns
    -------
    A SimulationRequest

    """  # noqa: E501
    if isinstance(simulation_input, ThermalHistoryInput):
        request = _setup_thermal_history(simulation_input, server, progress_handler)
    else:
        request = simulation_input._to_simulation_request()

    return request



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides input and result summary containers for single bead simulations."""

import contextlib
import math
import os
import zipfile

import numpy as np
from pandas import DataFrame

from ansys.additive.core.machine import AdditiveMachine
from ansys.additive.core.material import AdditiveMaterial
from ansys.additive.core.simulation import (
    SimulationInputBase,
    SimulationStatus,
    SimulationSummaryBase,
)
from ansys.api.additive.v0.additive_domain_pb2 import MeltPool as MeltPoolMessage
from ansys.api.additive.v0.additive_domain_pb2 import (
    SingleBeadInput as SingleBeadInputMessage,
)
from ansys.api.additive.v0.additive_simulation_pb2 import SimulationRequest


class SingleBeadInput(SimulationInputBase):
    """Provides input parameters for a single bead simulation."""

    DEFAULT_BEAD_LENGTH = 3e-3
    """Default bead length (m)."""
    MIN_BEAD_LENGTH = 1e-3
    """Minimum bead length (m)."""
    MAX_BEAD_LENGTH = 1e-2
    """Maximum bead length (m)."""
    DEFAULT_OUTPUT_THERMAL_HISTORY = False
    """Default output thermal history flag."""
    DEFAULT_THERMAL_HISTORY_INTERVAL = 1
    """Default thermal history interval."""
    MIN_THERMAL_HISTORY_INTERVAL = 1
    """Minimum thermal history interval."""
    MAX_THERMAL_HISTORY_INTERVAL = 10000
    """Maximum thermal history interval."""

    def __init__(
        self,
        *,
        bead_length: float = DEFAULT_BEAD_LENGTH,
        machine: AdditiveMachine = None,
        material: AdditiveMaterial = None,
        output_thermal_history: bool = DEFAULT_OUTPUT_THERMAL_HISTORY,
        thermal_history_interval: int = DEFAULT_THERMAL_HISTORY_INTERVAL,
    ):
        """Initialize a ``SingleBeadInput`` object."""
        super().__init__()
        self.bead_length = bead_length
        self.machine = machine if machine else AdditiveMachine()
        self.material = material if material else AdditiveMaterial()
        self.output_thermal_history = output_thermal_history
        self.thermal_history_interval = thermal_history_interval

    def __repr__(self):
        repr = type(self).__name__ + "\n"
        for k in self.__dict__:
            if k == "_machine" or k == "_material":
                repr += "\n" + k.replace("_", "", 1) + ": " + str(getattr(self, k))
            else:
                repr += k.replace("_", "", 1) + ": " + str(getattr(self, k)) + "\n"
        return repr

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, SingleBeadInput):
            return False
        return (
            self.id == __o.id
            and self.bead_length == __o.bead_length
            and self.machine == __o.machine
            and self.material == __o.material
            and self.output_thermal_history == __o.output_thermal_history
            and self.thermal_history_interval == __o.thermal_history_interval
        )

    def __validate_range(self, value, min, max, name):
        if math.isnan(value):
            raise ValueError("{} must be a number.".format(name))
        if value < min or value > max:
            raise ValueError("{} must be between {} and {}.".format(name, min, max))

    @property
    def machine(self) -> AdditiveMachine:
        """Machine parameters."""
        return self._machine

    @machine.setter
    def machine(self, value):
        self._machine = value

    @property
    def material(self) -> AdditiveMaterial:
        """Material parameters."""
        return self._material

    @material.setter
    def material(self, value):
        self._material = value

    @property
    def bead_length(self) -> float:
        """Length (m) of bead to simulate.

        Valid values are from the :obj:`MIN_BEAD_LENGTH` value to the
        :obj:`MAX_BEAD_LENGTH` value.
        """
        return self._bead_length

    @bead_length.setter
    def bead_length(self, value):
        self.__validate_range(value, self.MIN_BEAD_LENGTH, self.MAX_BEAD_LENGTH, "bead_length")
        self._bead_length = value

    @property
    def output_thermal_history(self) -> bool:
        """Flag indicating whether to output the thermal history of the simulation."""
        return self._output_thermal_history

    @output_thermal_history.setter
    def output_thermal_history(self, value: bool):
        self._output_thermal_history = value

    @property
    def thermal_history_interval(self) -> int:
        """Interval, in simulation steps, between thermal history results.

        Use ``1`` to create thermal history results for every simulation step,
        ``2`` for every other step, and so on.
        """
        return self._thermal_history_interval

    @thermal_history_interval.setter
    def thermal_history_interval(self, value: int):
        self.__validate_range(
            value,
            self.MIN_THERMAL_HISTORY_INTERVAL,
            self.MAX_THERMAL_HISTORY_INTERVAL,
            "thermal_history_interval",
        )
        self._thermal_history_interval = value

    def _to_simulation_request(self) -> SimulationRequest:
        """Convert this object into a simulation request message."""
        input = SingleBeadInputMessage(
            machine=self.machine._to_machine_message(),
            material=self.material._to_material_message(),
            bead_length=self.bead_length,
            output_thermal_history=self.output_thermal_history,
            thermal_history_interval=self.thermal_history_interval,
        )
        return SimulationRequest(id=self.id, single_bead_input=input)


class MeltPoolColumnNames:
    """Provides column names for the melt pool data frame."""

    WIDTH = "width"
    """Width of melt pool (m)."""
    DEPTH = "depth"
    """Depth of melt pool (m)."""
    LENGTH = "length"
    """Length of melt pool (m)."""
    REFERENCE_WIDTH = "reference_width"
    """Width of melt pool at the surface of the base plate (m)."""
    REFERENCE_DEPTH = "reference_depth"
    """Depth of melt pool measured from the surface of the base plate (m)."""


class MeltPool:
    """Contains the melt pool size dimensions for each time step during a single bead simulation."""

    def __init__(self, msg: MeltPoolMessage, thermal_history_output: str | None = None):
        """Initialize a ``MeltPool`` object.

        Parameters
        ----------
        msg: MeltPoolMessage
            The message containing the melt pool data.
        thermal_history_output: str | None
            Path to the thermal history output file.

        """
        bead_length = [ts.laser_x for ts in msg.time_steps]
        length = [ts.length for ts in msg.time_steps]
        width = [ts.width for ts in msg.time_steps]
        depth = [ts.depth for ts in msg.time_steps]
        reference_width = [ts.reference_width for ts in msg.time_steps]
        reference_depth = [ts.reference_depth for ts in msg.time_steps]
        self._df = DataFrame(
            index=bead_length,
            data={
                MeltPoolColumnNames.LENGTH: length,
                MeltPoolColumnNames.WIDTH: width,
                MeltPoolColumnNames.DEPTH: depth,
                MeltPoolColumnNames.REFERENCE_WIDTH: reference_width,
                MeltPoolColumnNames.REFERENCE_DEPTH: reference_depth,
            },
        )
        self._df.index.name = "bead_length"
        self._thermal_history_output = thermal_history_output

    def data_frame(self) -> DataFrame:
        """Get the data frame containing the melt pool data.

        Values are in meters.

        Indices:
            - ``bead_length``: Length of the bead at each time step.

        Columns:
            - :obj:`MeltPoolColumnNames.LENGTH`.
            - :obj:`MeltPoolColumnNames.WIDTH`.
            - :obj:`MeltPoolColumnNames.DEPTH`.
            - :obj:`MeltPoolColumnNames.REFERENCE_WIDTH`.
            - :obj:`MeltPoolColumnNames.REFERENCE_DEPTH`.
        """
        return self._df.copy()

    def depth_over_width(self) -> float:
        """Return the median reference depth over reference width."""
        depth = self.median_reference_depth()
        width = self.median_reference_width()
        return depth / width if width != 0 else np.nan

    def length_over_width(self) -> float:
        """Return the median length over width."""
        length = self.median_length()
        width = self.median_width()
        return length / width if width != 0 else np.nan

    def median_width(self) -> float:
        """Return the median width."""
        return self._df[MeltPoolColumnNames.WIDTH].median()

    def median_depth(self) -> float:
        """Return the median depth."""
        return self._df[MeltPoolColumnNames.DEPTH].median()

    def median_length(self) -> float:
        """Return the median length."""
        return self._df[MeltPoolColumnNames.LENGTH].median()

    def median_reference_width(self) -> float:
        """Return the median reference width."""
        return self._df[MeltPoolColumnNames.REFERENCE_WIDTH].median()

    def median_reference_depth(self) -> float:
        """Return the median reference depth."""
        return self._df[MeltPoolColumnNames.REFERENCE_DEPTH].median()

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, MeltPool):
            return False
        return self._df.equals(__o._df)

    def __repr__(self):
        repr = type(self).__name__ + "\n"
        repr += self._df.to_string()
        repr += "\ngrid_full_thermal_sensor_file_output_path: " + str(self.thermal_history_output)
        return repr

    @property
    def thermal_history_output(self) -> str | None:
        """Path to the thermal history output file."""
        return self._thermal_history_output


class SingleBeadSummary(SimulationSummaryBase):
    """Provides a summary of a single bead simulation."""

    THERMAL_HISTORY_OUTPUT_ZIP = "gridfullthermal.zip"

    def __init__(
        self,
        input: SingleBeadInput,
        msg: MeltPoolMessage,
        logs: str,
        thermal_history_output: str | None = None,
        status: SimulationStatus = SimulationStatus.COMPLETED,
    ):
        """Initialize a ``SingleBeadSummary`` object."""
        if not isinstance(input, SingleBeadInput):
            raise ValueError("Invalid input type passed to init, " + self.__class__.__name__)
        if not isinstance(msg, MeltPoolMessage):
            raise ValueError("Invalid message type passed to init, " + self.__class__.__name__)
        if not isinstance(logs, str):
            raise ValueError("Invalid logs type passed to init, " + self.__class__.__name__)
        super().__init__(logs, status)
        self._input = input
        self._melt_pool = MeltPool(msg, thermal_history_output)
        if thermal_history_output is not None:
            self._extract_thermal_history(thermal_history_output)

    @property
    def input(self) -> SingleBeadInput:
        """Simulation input."""
        return self._input

    @property
    def melt_pool(self) -> MeltPool:
        """Resulting melt pool."""
        return self._melt_pool

    def __repr__(self):
        repr = type(self).__name__ + "\n"
        for k in self.__dict__:
            repr += k.replace("_", "", 1) + ": " + str(getattr(self, k)) + "\n"
        return repr

    def _extract_thermal_history(self, thermal_history_output):
        """Extract the thermal history output."""
        zip_file = os.path.join(thermal_history_output, self.THERMAL_HISTORY_OUTPUT_ZIP)
        if not os.path.isfile(zip_file):
            raise FileNotFoundError("Thermal history files not found: " + zip_file)
        with zipfile.ZipFile(zip_file, "r") as zip_ref:
            zip_ref.extractall(thermal_history_output)
        with contextlib.suppress(OSError):
            os.remove(zip_file)



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides a client for interacting with the Additive service."""

import logging
import os
import time

import grpc
from google.longrunning.operations_pb2 import Operation
from google.protobuf.empty_pb2 import Empty

import ansys.additive.core.misc as misc
from ansys.additive.core import USER_DATA_PATH, __version__
from ansys.additive.core.download import download_logs
from ansys.additive.core.exceptions import BetaFeatureNotEnabledError
from ansys.additive.core.logger import LOG
from ansys.additive.core.material import RESERVED_MATERIAL_NAMES, AdditiveMaterial
from ansys.additive.core.material_tuning import (
    MaterialTuningInput,
    MaterialTuningSummary,
)
from ansys.additive.core.microstructure import (
    MicrostructureInput,
    MicrostructureSummary,
)
from ansys.additive.core.microstructure_3d import (
    Microstructure3DInput,
    Microstructure3DSummary,
)
from ansys.additive.core.parametric_study import ParametricStudy
from ansys.additive.core.parametric_study.parametric_study_progress_handler import (
    ParametricStudyProgressHandler,
)
from ansys.additive.core.porosity import PorosityInput, PorositySummary
from ansys.additive.core.progress_handler import (
    DefaultSingleSimulationProgressHandler,
    IProgressHandler,
)
from ansys.additive.core.server_connection import (
    DEFAULT_PRODUCT_VERSION,
    ServerConnection,
)
from ansys.additive.core.simulation import (
    SimulationStatus,
    SimulationType,
)
from ansys.additive.core.simulation_error import SimulationError
from ansys.additive.core.simulation_requests import create_request
from ansys.additive.core.simulation_task import SimulationTask
from ansys.additive.core.simulation_task_manager import SimulationTaskManager
from ansys.additive.core.single_bead import SingleBeadInput, SingleBeadSummary
from ansys.additive.core.thermal_history import (
    ThermalHistoryInput,
    ThermalHistorySummary,
)
from ansys.api.additive import __version__ as api_version
from ansys.api.additive.v0.additive_materials_pb2 import (
    AddMaterialRequest,
    GetMaterialRequest,
    RemoveMaterialRequest,
)
from ansys.api.additive.v0.additive_operations_pb2 import OperationMetadata
from ansys.api.additive.v0.additive_settings_pb2 import SettingsRequest


class Additive:
    """Provides the client interface to one or more Additive services.

    Parameters
    ----------
    channel: grpc.Channel, default: None
        Server connection. If provided, it is assumed that the
        :class:`grpc.Channel <grpc.Channel>` object is connected to the server.
        Also, if provided, the ``host`` and ``port`` parameters are ignored.
    host: str, default: None
        Host name or IPv4 address of the server. This parameter is ignored if the
        ``server_channels`` or ``channel`` parameters is other than ``None``.
    port: int, default: 50052
        Port number to use when connecting to the server.
    nsims_per_server: int, default: 1
        Number of simultaneous simulations to run on the server. Each simulation
        requires a license checkout. If a license is not available, the simulation
        fails.
    product_version: str
        Version of the Ansys product installation in the form ``"YYR"``, where ``YY``
        is the two-digit year and ``R`` is the release number. For example, the release
        2024 R1 would be specified as ``241``. This parameter is only applicable in
        `PyPIM`_-enabled cloud environments and on localhost. Using an empty string
        or ``None`` uses the default product version.
    log_level: str, default: ""
        Minimum severity level of messages to log. Valid values are "DEBUG", "INFO",
        "WARNING", "ERROR", and "CRITICAL". The default value equates to "WARNING".
    log_file: str, default: ""
        File name to write log messages to.
    enable_beta_features: bool, default: False
        Flag indicating if beta features are enabled.
    linux_install_path: os.PathLike, None, default: None
        Path to the Ansys installation directory on Linux. This parameter is only
        required when Ansys has not been installed in the default location. Example:
        ``/usr/shared/ansys_inc``. Note that the path should not include the product
        version.

    Examples
    --------
    Connect to a list of servers. Multiple connections to the same host are permitted.

    >>> additive = Additive(server_connections=["localhost:50052", "localhost:50052", "myserver:50052"])

    Connect to a single server using the host name and port number.

    >>> additive = Additive(host="additive.ansys.com", port=12345)

    Start and connect to a server on localhost or in a
    `PyPIM`_-enabled cloud environment. Allow two simultaneous
    simulations on the server.

    >>> additive = Additive(nsims_per_server=2)

    Start a single server on localhost or in a `PyPIM`_-enabled cloud environment.
    Use version 2024 R1 of the Ansys product installation.

    >>> additive = Additive(product_version="241")

    .. _PyPIM: https://pypim.docs.pyansys.com/version/stable/index.html

    """

    DEFAULT_ADDITIVE_SERVICE_PORT = 50052

    def __init__(
        self,
        channel: grpc.Channel | None = None,
        host: str | None = None,
        port: int = DEFAULT_ADDITIVE_SERVICE_PORT,
        nsims_per_server: int = 1,
        product_version: str = DEFAULT_PRODUCT_VERSION,
        log_level: str = "",
        log_file: str = "",
        enable_beta_features: bool = False,
        linux_install_path: os.PathLike | None = None,
    ) -> None:
        """Initialize server connections."""
        if not product_version:
            product_version = DEFAULT_PRODUCT_VERSION

        if log_level:
            LOG.setLevel(log_level)
        if log_file:
            LOG.log_to_file(filename=log_file, level=log_level)

        self._server = Additive._connect_to_server(
            channel,
            host,
            port,
            product_version,
            LOG,
            linux_install_path,
        )

        # HACK: Set the number of concurrent simulations per server
        # when generating documentation to reduce time.
        if os.getenv("GENERATING_DOCS", None):
            nsims_per_server = 8
        initial_settings = {"NumConcurrentSims": str(nsims_per_server)}
        LOG.info(self.apply_server_settings(initial_settings))

        self._enable_beta_features = enable_beta_features

        # Setup data directory
        self._user_data_path = USER_DATA_PATH
        if not os.path.exists(self._user_data_path):  # pragma: no cover
            os.makedirs(self._user_data_path)
        LOG.info("user data path: " + self._user_data_path)

    @staticmethod
    def _connect_to_server(
        channel: grpc.Channel | None = None,
        host: str | None = None,
        port: int = DEFAULT_ADDITIVE_SERVICE_PORT,
        product_version: str = DEFAULT_PRODUCT_VERSION,
        log: logging.Logger = None,
        linux_install_path: os.PathLike | None = None,
    ) -> ServerConnection:
        """Connect to an Additive server, starting it if necessary.

        Parameters
        ----------
        channel: grpc.Channel, default: None
            Server connection. If provided, it is assumed to be connected
            and the ``host`` and ``port`` parameters are ignored.
        host: str, default: None
            Host name or IPv4 address of the server. This parameter is ignored if
            the ``channel`` parameter is other than ``None``.
        port: int, default: 50052
            Port number to use when connecting to the server.
        product_version: str
            Version of the Ansys product installation in the form ``"YYR"``, where ``YY``
            is the two-digit year and ``R`` is the release number. For example, "251".
            This parameter is only applicable in `PyPIM`_-enabled cloud environments and
            on localhost. Using an empty string or ``None`` uses the default product version.
        log: logging.Logger, default: None
            Logger to use for logging messages.
        linux_install_path: os.PathLike, None, default: None
            Path to the Ansys installation directory on Linux. This parameter is only
            required when Ansys has not been installed in the default location. Example:
            ``/usr/shared/ansys_inc``. Note that the path should not include the product
            version.

        Returns
        -------
        ServerConnection
            Connection to the server.

        NOTE: If ``channel`` and ``host`` are not provided and the environment variable
        ``ANSYS_ADDITIVE_ADDRESS`` is set, the client will connect to the server at the
        address specified by the environment variable. The value of the environment variable
        should be in the form ``host:port``.

        """
        if channel:
            if not isinstance(channel, grpc.Channel):
                raise ValueError("channel must be a grpc.Channel object")
            return ServerConnection(channel=channel, log=log)
        elif host:
            return ServerConnection(addr=f"{host}:{port}", log=log)
        elif os.getenv("ANSYS_ADDITIVE_ADDRESS"):
            return ServerConnection(addr=os.getenv("ANSYS_ADDITIVE_ADDRESS"), log=log)
        else:
            return ServerConnection(
                product_version=product_version,
                log=log,
                linux_install_path=linux_install_path,
            )

    @property
    def enable_beta_features(self) -> bool:
        """Flag indicating if beta features are enabled."""
        return self._enable_beta_features

    @enable_beta_features.setter
    def enable_beta_features(self, value: bool) -> None:
        """Set the flag indicating if beta features are enabled."""
        self._enable_beta_features = value

    @property
    def connected(self) -> bool:
        """Return True if the client is connected to a server."""
        if self._server is None:
            return False
        return self._server.status().connected

    def about(self) -> str:
        """Return information about the client and server.

        Returns
        -------
        str
            Information about the client and server.

        """
        about = (
            f"ansys.additive.core version {__version__}\nClient side API version: {api_version}\n"
        )
        if self._server is None:
            about += "Client is not connected to a server.\n"
        else:
            about += str(self._server.status()) + "\n"
        return about

    def apply_server_settings(self, settings: dict[str, str]) -> list[str]:
        """Apply settings to each server.

        Current settings include:
        - ``NumConcurrentSims``: number of concurrent simulations per server.

        Parameters
        ----------
        settings: dict[str, str]
            Dictionary of settings to apply to the server.

        Returns
        -------
        list[str]
            List of messages from the server.

        """
        request = SettingsRequest()
        for setting_key, setting_value in settings.items():
            setting = request.settings.add()
            setting.key = setting_key
            setting.value = setting_value

        response = self._server.settings_stub.ApplySettings(request)

        return response.messages

    def list_server_settings(self) -> dict[str, str]:
        """Get a dictionary of settings for the server."""
        response = self._server.settings_stub.ListSettings(Empty())
        settings = {}
        for setting in response.settings:
            settings[setting.key] = setting.value
        return settings

    def simulate(
        self,
        inputs: (
            SingleBeadInput
            | PorosityInput
            | MicrostructureInput
            | ThermalHistoryInput
            | Microstructure3DInput
            | list
        ),
        progress_handler: IProgressHandler | None = None,
    ) -> (
        SingleBeadSummary
        | PorositySummary
        | MicrostructureSummary
        | ThermalHistorySummary
        | Microstructure3DSummary
        | SimulationError
        | list
    ):
        """Execute additive simulations.

        Parameters
        ----------
        inputs: SingleBeadInput, PorosityInput, MicrostructureInput, ThermalHistoryInput, Microstructure3DInput, list
            Parameters to use for simulations. A list of inputs may be provided to run multiple
            simulations.
        progress_handler: IProgressHandler, None, default: None
            Handler for progress updates. If ``None``, and ``inputs`` contains a single
            simulation input, a default progress handler will be assigned.

        Returns
        -------
        SingleBeadSummary, PorositySummary, MicrostructureSummary, ThermalHistorySummary,
        Microstructure3DSummary, SimulationError, list
            One or more summaries of simulation results. If a list of inputs is provided, a
            list is returned.

        """  # noqa: E501
        summaries = []
        task_mgr = self.simulate_async(inputs, progress_handler)
        task_mgr.wait_all(progress_handler=progress_handler)
        summaries = task_mgr.summaries()

        for summ in summaries:
            if isinstance(summ, SimulationError):
                LOG.error(f"\nError: {summ.message}")

        return summaries if isinstance(inputs, list) else summaries[0]

    def simulate_async(
        self,
        inputs: (
            SingleBeadInput
            | PorosityInput
            | MicrostructureInput
            | ThermalHistoryInput
            | Microstructure3DInput
            | list
        ),
        progress_handler: IProgressHandler | None = None,
    ) -> SimulationTaskManager:
        """Execute additive simulations asynchronously. This method does not block while the
        simulations are running on the server. This class stores handles of type
        google.longrunning.Operation to the remote tasks that can be used to communicate with
        the server for status updates.

        Parameters
        ----------
        inputs: SingleBeadInput, PorosityInput, MicrostructureInput, ThermalHistoryInput, Microstructure3DInput, list
            Parameters to use for simulations. A list of inputs may be provided to run multiple
            simulations.
        progress_handler: IProgressHandler, None, default: None
            Handler for progress updates. If ``None``, and ``inputs`` contains a single
            simulation input, a default progress handler will be assigned.

        Returns
        -------
        SimulationTaskManager
            A SimulationTaskManager to handle all tasks sent to the server by this function call.

        """  # noqa: E501
        self._check_for_duplicate_id(inputs)

        task_manager = SimulationTaskManager()
        if not isinstance(inputs, list):
            if not progress_handler:
                progress_handler = DefaultSingleSimulationProgressHandler()
            simulation_task = self._simulate(inputs, self._server, progress_handler)
            task_manager.add_task(simulation_task)
            return task_manager

        if len(inputs) == 0:
            raise ValueError("No simulation inputs provided")

        LOG.info(f"Starting {len(inputs)} simulations")
        for sim_input in inputs:
            task = self._simulate(sim_input, self._server, progress_handler)
            task_manager.add_task(task)

        return task_manager

    def _simulate(
        self,
        simulation_input: (
            SingleBeadInput
            | PorosityInput
            | MicrostructureInput
            | ThermalHistoryInput
            | Microstructure3DInput
        ),
        server: ServerConnection,
        progress_handler: IProgressHandler | None = None,
    ) -> SimulationTask:
        """Execute a single simulation.

        Parameters
        ----------
        simulation_input: SingleBeadInput, PorosityInput, MicrostructureInput, ThermalHistoryInput, Microstructure3DInput
            Parameters to use for simulation.
        server: ServerConnection
            Server to use for the simulation.
        progress_handler: IProgressHandler, None, default: None
            Handler for progress updates. If ``None``, no progress updates are provided.

        Returns
        -------
        SimulationTask
            A task that can be used to monitor the simulation progress.

        """  # noqa: E501

        if simulation_input.material == AdditiveMaterial():
            raise ValueError("A material is not assigned to the simulation input")

        if (
            isinstance(simulation_input, (Microstructure3DInput, ThermalHistoryInput))
            and self.enable_beta_features is False
        ):
            raise BetaFeatureNotEnabledError(
                "This simulation requires beta features to be enabled.\n"
                "Set enable_beta_features=True when creating the Additive client."
            )

        try:
            request = create_request(simulation_input, server, progress_handler)
            long_running_op = server.simulation_stub.Simulate(request)
            simulation_task = SimulationTask(
                server, long_running_op, simulation_input, self._user_data_path
            )
            LOG.debug(f"Simulation task created for {simulation_input.id}")

        except Exception as e:
            metadata = OperationMetadata(simulation_id=simulation_input.id, message=str(e))
            errored_op = Operation(name=simulation_input.id, done=True)
            errored_op.metadata.Pack(metadata)
            simulation_task = SimulationTask(
                server, errored_op, simulation_input, self._user_data_path
            )

        if progress_handler:
            time.sleep(0.1)  # Allow time for the server to start the simulation
            progress_handler.update(simulation_task.status())

        return simulation_task

    def materials_list(self) -> list[str]:
        """Get a list of material names used in additive simulations.

        Returns
        -------
        list[str]
            Names of available additive materials.

        """
        response = self._server.materials_stub.GetMaterialsList(Empty())
        return response.names

    def material(self, name: str) -> AdditiveMaterial:
        """Get a material for use in an additive simulation.

        Parameters
        ----------
        name: str
            Name of material.

        Returns
        -------
        AdditiveMaterial
            Requested material definition.

        """
        request = GetMaterialRequest(name=name)
        result = self._server.materials_stub.GetMaterial(request)
        return AdditiveMaterial._from_material_message(result)

    @staticmethod
    def load_material(
        parameters_file: str,
        thermal_lookup_file: str,
        characteristic_width_lookup_file: str,
    ) -> AdditiveMaterial:
        """Load a custom material definition for the current session. The resulting
        ``AdditiveMaterial`` object will not be saved to the library.

        Parameters
        ----------
        parameters_file: str
            Name of the JSON file containing material parameters. For more information, see
            `Create Material Configuration File (.json)
            <https://ansyshelp.ansys.com/account/secured?returnurl=/Views/Secured/corp/v232/en/add_beta/add_print_udm_tool_create_tables.html>`_
            in the *Additive Manufacturing Beta Features* documentation.
        thermal_lookup_file: str
            Name of the CSV file containing the lookup table for thermal dependent properties.
            For more information, see `Create Material Lookup File (.csv)
            <https://ansyshelp.ansys.com/account/secured?returnurl=/Views/Secured/corp/v232/en/add_beta/add_print_udm_create_mat_lookup.html>`_
            in the *Additive Manufacturing Beta Features* documentation.
        characteristic_width_lookup_file: str
            Name of the CSV file containing the lookup table for characteristic melt pool width. For
            more information, see
            `Find Characteristic Width Values and Generate Characteristic Width File (.csv)
            <https://ansyshelp.ansys.com/account/secured?returnurl=/Views/Secured/corp/v232/en/add_beta/add_print_udm_tool_find_cw.html>`_
            in the *Additive Manufacturing Beta Features* documentation.

        Returns
        -------
        AdditiveMaterial
            A material definition for use in additive simulations.

        """
        material = AdditiveMaterial()
        material._load_parameters(parameters_file)
        material._load_thermal_properties(thermal_lookup_file)
        material._load_characteristic_width(characteristic_width_lookup_file)
        return material

    def add_material(
        self,
        parameters_file: str,
        thermal_lookup_file: str,
        characteristic_width_lookup_file: str,
    ) -> AdditiveMaterial | None:
        """Add a custom material to the library for use in additive simulations.

        Parameters
        ----------
        parameters_file: str
            Name of the JSON file containing material parameters. For more information, see
            `Create Material Configuration File (.json)
            <https://ansyshelp.ansys.com/account/secured?returnurl=/Views/Secured/corp/v232/en/add_beta/add_print_udm_tool_create_tables.html>`_
            in the *Additive Manufacturing Beta Features* documentation.
        thermal_lookup_file: str
            Name of the CSV file containing the lookup table for temperature-dependent properties.
            For more information, see `Create Material Lookup File (.csv)
            <https://ansyshelp.ansys.com/account/secured?returnurl=/Views/Secured/corp/v232/en/add_beta/add_print_udm_create_mat_lookup.html>`_
            in the *Additive Manufacturing Beta Features* documentation.
        characteristic_width_lookup_file: str
            Name of the CSV file containing the lookup table for characteristic melt pool width. For
            more information, see
            `Find Characteristic Width Values and Generate Characteristic Width File (.csv)
            <https://ansyshelp.ansys.com/account/secured?returnurl=/Views/Secured/corp/v232/en/add_beta/add_print_udm_tool_find_cw.html>`_
            in the *Additive Manufacturing Beta Features* documentation.

        Returns
        -------
        AdditiveMaterial
            A definition of the material that was added to the library.

        """  # noqa: E501
        material = self.load_material(
            parameters_file, thermal_lookup_file, characteristic_width_lookup_file
        )

        names = self.materials_list()
        if material.name.lower() in (name.lower() for name in names):
            raise ValueError(f"Material {material.name} already exists. Unable to add material.")

        request = AddMaterialRequest(id=misc.short_uuid(), material=material._to_material_message())
        LOG.info(f"Adding material {request.material.name}")
        response = self._server.materials_stub.AddMaterial(request)

        if response.HasField("error"):
            raise RuntimeError(response.error)

        return AdditiveMaterial._from_material_message(response.material)

    def remove_material(self, name: str):
        """Remove a material from the server.

        Parameters
        ----------
        name: str
            Name of the material to remove.

        """
        if name.lower() in (material.lower() for material in RESERVED_MATERIAL_NAMES):
            raise ValueError(f"Unable to remove Ansys-supplied material '{name}'.")

        self._server.materials_stub.RemoveMaterial(RemoveMaterialRequest(name=name))

    def tune_material(
        self,
        input: MaterialTuningInput,
        out_dir: str = USER_DATA_PATH,
        progress_handler: IProgressHandler = None,
    ) -> MaterialTuningSummary | None:
        """Tune a custom material for use with additive simulations.

        This method performs the same function as the Material Tuning Tool
        described in
        `Find Simulation Parameters to Match Simulation to Experiments
        <https://ansyshelp.ansys.com/account/secured?returnurl=/Views/Secured/corp/v232/en/add_beta/add_print_udm_tool_match_sim_to_exp.html>`_.
        It is used for one step in the material tuning process. The other steps
        are described in
        `Chapter 2: Material Tuning Tool (Beta) to Create User Defined Materials
        <https://ansyshelp.ansys.com/account/secured?returnurl=/Views/Secured/corp/v232/en/add_beta/add_science_BETA_material_tuning_tool.html>`_.

        Parameters
        ----------
        input: MaterialTuningInput
            Input parameters for material tuning.
        out_dir: str, default: USER_DATA_PATH
            Folder path for output files.
        progress_handler: IProgressHandler, None, default: None
            Handler for progress updates. If ``None``, no progress updates are provided.

        Returns
        -------
        MaterialTuningSummary, None
            Summary of material tuning or ``None`` if the tuning failed.

        """  # noqa: E501

        task = self.tune_material_async(input, out_dir)
        task.wait(progress_handler=progress_handler)
        return task.summary

    def tune_material_async(
        self,
        input: MaterialTuningInput,
        out_dir: str = USER_DATA_PATH,
    ) -> SimulationTask:
        """Tune a custom material for use with additive simulations asynchronously.

        This method performs the same function as the Material Tuning Tool
        described in
        `Find Simulation Parameters to Match Simulation to Experiments
        <https://ansyshelp.ansys.com/account/secured?returnurl=/Views/Secured/corp/v232/en/add_beta/add_print_udm_tool_match_sim_to_exp.html>`_.
        It is used for one step in the material tuning process. The other steps
        are described in
        `Chapter 2: Material Tuning Tool (Beta) to Create User Defined Materials
        <https://ansyshelp.ansys.com/account/secured?returnurl=/Views/Secured/corp/v232/en/add_beta/add_science_BETA_material_tuning_tool.html>`_.

        Parameters
        ----------
        input: MaterialTuningInput
            Input parameters for material tuning.
        out_dir: str, default: USER_DATA_PATH
            Folder path for output files.

        Returns
        -------
        SimulationTask
            An asynchronous simulation task.

        """
        if input.id == "":
            input.id = misc.short_uuid()
        if out_dir == USER_DATA_PATH:
            out_dir = os.path.join(out_dir, input.id)

        if os.path.exists(out_dir):
            raise ValueError(
                f"Directory {out_dir} already exists. Delete or choose a different output directory."
            )

        request = input._to_request()
        operation = self._server.simulation_stub.Simulate(request)
        LOG.debug(f"Material tuning operation created for {input.id}")
        return SimulationTask(self._server, operation, input, out_dir)

    def simulate_study(
        self,
        study: ParametricStudy,
        simulation_ids: list[str] | None = None,
        types: list[SimulationType] | None = None,
        priority: int | None = None,
        iteration: int = None,
    ):
        """Run the simulations in a parametric study.

        Parameters
        ----------
        study : ParametricStudy
            Parametric study to run.
        simulation_ids : list[str], default: None
            List of simulation IDs to run. If this value is ``None``,
            all simulations with a status of ``Pending`` are run.
        types : list[SimulationType], default: None
            Type of simulations to run. If this value is ``None``,
            all simulation types are run.
        priority : int, default: None
            Priority of simulations to run. If this value is ``None``,
            all priorities are run.
        iteration : int, default: None
            Iteration number of simulations to run. The default is ``None``,
            all iterations are run.
        progress_handler : IProgressHandler, None, default: None
            Handler for progress updates. If ``None``, a :class:`ParametricStudyProcessHandler` will be used.

        """
        SLEEP_INTERVAL = 2
        progress_handler = ParametricStudyProgressHandler(study)
        summaries = []

        try:
            task_mgr = self.simulate_study_async(
                study, simulation_ids, types, priority, iteration, progress_handler
            )
            # Allow time for the server to start the simulations
            time.sleep(SLEEP_INTERVAL)
            while not task_mgr.done:
                task_mgr.status(progress_handler)
                current_summaries = task_mgr.summaries()
                new_summaries = [s for s in current_summaries if s not in summaries]
                if new_summaries:
                    study.update(new_summaries)
                    summaries = current_summaries
                time.sleep(SLEEP_INTERVAL)

        except Exception as e:
            LOG.error(f"Error running study: {e}")
            study.reset_simulation_status()
            raise RuntimeError from e

    def simulate_study_async(
        self,
        study: ParametricStudy,
        simulation_ids: list[str] | None = None,
        types: list[SimulationType] | None = None,
        priority: int | None = None,
        iteration: int = None,
        progress_handler: IProgressHandler = None,
    ) -> SimulationTaskManager:
        """Run the simulations in a parametric study asynchronously.

        Notes
        -----
            The caller of this method is responsible for updating the study with the results of the simulations.
            See :meth:`SimulationTaskManager.summaries` and :meth:`ParametricStudy.update`.

        Parameters
        ----------
        study : ParametricStudy
            Parametric study to run.
        simulation_ids : list[str], default: None
            List of simulation IDs to run. If this value is ``None``,
            all simulations with a status of ``Pending`` are run.
        types : list[SimulationType], default: None
            Type of simulations to run. If this value is ``None``,
            all simulation types are run.
        priority : int, default: None
            Priority of simulations to run. If this value is ``None``,
            all priorities are run.
        iteration : int, default: None
            Iteration number of simulations to run. The default is ``None``,
            all iterations are run.
        progress_handler : IProgressHandler, None, default: None
            Handler for progress updates.

        """
        inputs = study.simulation_inputs(self.material, simulation_ids, types, priority, iteration)
        if not inputs:
            # no simulations met the provided criteria, return an empty task manager
            return SimulationTaskManager()
        ids = [i.id for i in inputs]
        study.set_simulation_status(ids, SimulationStatus.PENDING)
        study.clear_errors(ids)
        return self.simulate_async(inputs, progress_handler)

    def _check_for_duplicate_id(self, inputs):
        """Check for duplicate simulation IDs in a list of inputs.

        If an input does not have an ID, one will be assigned.

        Parameters
        ----------
        inputs : SingleBeadInput, PorosityInput, MicrostructureInput, ThermalHistoryInput, Microstructure3DInput, list
            A simulation input or a list of simulation inputs.

        Raises
        ------
        ValueError
            If a duplicate ID is found in the list of inputs.

        """  # noqa: E501
        if not isinstance(inputs, list):
            # An individual input, not a list
            if inputs.id == "":
                inputs.id = misc.short_uuid()
            return
        ids = []
        for i in inputs:
            if not i.id:
                # give input an id if none provided
                i.id = misc.short_uuid()
            if any(id == i.id for id in ids):
                raise ValueError(f'Duplicate simulation ID "{i.id}" in input list')
            ids.append(i.id)

    def download_server_logs(self, log_dir: str | os.PathLike) -> str:
        """Download server logs to a specified directory.

        Parameters
        ----------
        log_dir : str
            Directory to save the logs to.

        Returns
        -------
        str
            Path to the downloaded logs.

        """
        if log_dir and not os.path.exists(log_dir):
            os.makedirs(log_dir)
        return download_logs(self._server.simulation_stub, log_dir)
